"""
API for halo_finding



"""

# -----------------------------------------------------------------------------
# Copyright (c) yt Development Team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

import warnings

from numpy import VisibleDeprecationWarning

warnings.warn(
    "Running the HoP and FoF halo finders directly is deprecated. "
    + "Please run using the HaloCatalog. See "
    + "https://yt-astro-analysis.readthedocs.io for more information.",
    VisibleDeprecationWarning,
    stacklevel=2,
)

from yt_astro_analysis.halo_finding.halo_objects import FOFHaloFinder, HOPHaloFinder



"""
API for halo_analysis



"""

import warnings

from numpy import VisibleDeprecationWarning

from yt_astro_analysis.halo_analysis.halo_catalog.analysis_operators import add_quantity
from yt_astro_analysis.halo_analysis.halo_catalog.halo_callbacks import add_callback
from yt_astro_analysis.halo_analysis.halo_catalog.halo_catalog import HaloCatalog
from yt_astro_analysis.halo_analysis.halo_catalog.halo_filters import add_filter
from yt_astro_analysis.halo_analysis.halo_catalog.halo_finding_methods import (
    add_finding_method,
)
from yt_astro_analysis.halo_analysis.halo_catalog.halo_recipes import add_recipe

warnings.warn(
    "Importing from yt_astro_analysis.halo_analysis.api is deprecated. "
    + "Please import from yt_astro_analysis.halo_analysis or "
    + "yt.extensions.astro_analysis.halo_analysis.",
    VisibleDeprecationWarning,
    stacklevel=2,
)



"""
Halo object.



"""


class Halo:
    def __init__(self, halo_catalog, data_source, index):
        self.halo_catalog = halo_catalog
        self.data_source = data_source
        self.index = index
        self.quantities = {}

    def _get_field_value(self, fieldname):
        return self.data_source[fieldname][self.index]

    def _set_field_value(self, fieldkey, fieldname):
        self.quantities[fieldkey] = self._get_field_value(fieldname)



"""
Create randomly centered, tiled projections to be used in light cones.



"""

# -----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

import numpy as np

from yt.funcs import mylog
from yt.units.yt_array import uconcatenate
from yt.utilities.parallel_tools.parallel_analysis_interface import (
    parallel_blocking_call,
)
from yt.visualization.fixed_resolution import FixedResolutionBuffer


@parallel_blocking_call
def _light_cone_projection(
    my_slice, field, pixels, weight_field=None, save_image=False, field_cuts=None
):
    "Create a single projection to be added into the light cone stack."

    # We are just saving the projection object, so only the projection axis
    # needs to be considered since the lateral shifting and tiling occurs after
    # the projection object is made.
    # Likewise, only the box_depth_fraction needs to be considered.

    mylog.info(
        "Making projection at z = %f from %s.",
        my_slice["redshift"],
        my_slice["filename"],
    )

    region_center = [
        0.5
        * (
            my_slice["object"].domain_right_edge[q]
            + my_slice["object"].domain_left_edge[q]
        )
        for q in range(my_slice["object"].dimensionality)
    ]

    # 1. The Depth Problem
    # Use coordinate field cut in line of sight to cut projection to proper depth.
    if field_cuts is None:
        these_field_cuts = []
    else:
        these_field_cuts = field_cuts.copy()

    if my_slice["box_depth_fraction"] < 1:
        axis = ("x", "y", "z")[my_slice["projection_axis"]]
        depthLeft = (
            my_slice["projection_center"][my_slice["projection_axis"]]
            - 0.5 * my_slice["box_depth_fraction"]
        )
        depthRight = (
            my_slice["projection_center"][my_slice["projection_axis"]]
            + 0.5 * my_slice["box_depth_fraction"]
        )
        if depthLeft < 0:
            cut_mask = (
                "((obj['index', '%s'] + 0.5*obj['index', 'd%s'] >= 0) & "
                " (obj['index', '%s'] - 0.5*obj['index', 'd%s'] <= %f)) | "
                "((obj['index', '%s'] + 0.5*obj['index', 'd%s'] >= %f) & "
                " (obj['index', '%s'] - 0.5*obj['index', 'd%s'] <= 1))"
            ) % (
                axis,
                axis,
                axis,
                axis,
                depthRight,
                axis,
                axis,
                (depthLeft + 1),
                axis,
                axis,
            )
        elif depthRight > 1:
            cut_mask = (
                "((obj['index', '%s'] + 0.5*obj['index', 'd%s'] >= 0) & "
                "(obj['index', '%s'] - 0.5*obj['index', 'd%s'] <= %f)) | "
                "((obj['index', '%s'] + 0.5*obj['index', 'd%s'] >= %f) & "
                "(obj['index', '%s'] - 0.5*obj['index', 'd%s'] <= 1))"
            ) % (
                axis,
                axis,
                axis,
                axis,
                (depthRight - 1),
                axis,
                axis,
                depthLeft,
                axis,
                axis,
            )
        else:
            cut_mask = (
                "(obj['index', '%s'] + 0.5*obj['index', 'd%s'] >= %f) & "
                "(obj['index', '%s'] - 0.5*obj['index', '%s'] <= %f)"
            ) % (axis, axis, depthLeft, axis, axis, depthRight)

        these_field_cuts.append(cut_mask)

    data_source = my_slice["object"].all_data()
    cut_region = data_source.cut_region(these_field_cuts)

    # Make projection.
    proj = my_slice["object"].proj(
        field,
        my_slice["projection_axis"],
        weight_field=weight_field,
        center=region_center,
        data_source=cut_region,
    )
    proj_field = proj.field[0]

    if weight_field is not None:
        proj_weight_field = proj.weight_field
        # Do a projection of the weight field so we can keep it.
        weight_proj = my_slice["object"].proj(
            proj_weight_field,
            my_slice["projection_axis"],
            weight_field=None,
            center=region_center,
            data_source=cut_region,
        )

    del data_source, cut_region

    # 2. The Tile Problem
    # Tile projection to specified width.

    # Original projection data.
    original_px = proj["px"].in_units("code_length").copy()
    original_py = proj["py"].in_units("code_length").copy()
    original_pdx = proj["pdx"].in_units("code_length").copy()
    original_pdy = proj["pdy"].in_units("code_length").copy()
    original_field = proj[proj_field].copy()
    if weight_field is not None:
        original_weight_field = weight_proj[proj_weight_field].copy()

    for my_field in ["px", "py", "pdx", "pdy", proj_field]:
        proj.field_data[my_field] = [proj.field_data[my_field]]
    if weight_field is not None:
        weight_proj.field_data[proj_weight_field] = [
            weight_proj.field_data[proj_weight_field]
        ]

    # Copy original into offset positions to make tiles.
    for x in range(int(np.ceil(my_slice["box_width_fraction"]))):
        x = my_slice["object"].quan(x, "code_length")
        for y in range(int(np.ceil(my_slice["box_width_fraction"]))):
            y = my_slice["object"].quan(y, "code_length")
            if (x + y) > 0:
                proj.field_data["px"] += [original_px + x]
                proj.field_data["py"] += [original_py + y]
                proj.field_data["pdx"] += [original_pdx]
                proj.field_data["pdy"] += [original_pdy]
                proj.field_data[proj_field] += [original_field]
                if weight_field is not None:
                    weight_proj.field_data[proj_weight_field] += [original_weight_field]

    for my_field in ["px", "py", "pdx", "pdy", proj_field]:
        proj.field_data[my_field] = (
            my_slice["object"].arr(proj.field_data[my_field]).flatten()
        )
    if weight_field is not None:
        weight_proj.field_data[proj_weight_field] = (
            my_slice["object"].arr(weight_proj.field_data[proj_weight_field]).flatten()
        )

    # Delete originals.
    del original_px
    del original_py
    del original_pdx
    del original_pdy
    del original_field
    if weight_field is not None:
        del original_weight_field

    # 3. The Shift Problem
    # Shift projection by random x and y offsets.

    image_axes = np.roll(np.arange(3), -my_slice["projection_axis"])[1:]
    di_left_x = my_slice["object"].domain_left_edge[image_axes[0]]
    di_right_x = my_slice["object"].domain_right_edge[image_axes[0]]
    di_left_y = my_slice["object"].domain_left_edge[image_axes[1]]
    di_right_y = my_slice["object"].domain_right_edge[image_axes[1]]

    offset = my_slice["projection_center"].copy() * my_slice["object"].domain_width
    offset = np.roll(offset, -my_slice["projection_axis"])[1:]

    # Shift x and y positions.
    proj.field_data["px"] -= offset[0]
    proj.field_data["py"] -= offset[1]

    # Wrap off-edge cells back around to other side (periodic boundary conditions).
    proj.field_data["px"][proj.field_data["px"] < di_left_x] += (
        np.ceil(my_slice["box_width_fraction"]) * di_right_x
    )
    proj.field_data["py"][proj.field_data["py"] < di_left_y] += (
        np.ceil(my_slice["box_width_fraction"]) * di_right_y
    )

    # After shifting, some cells have fractional coverage on both sides of the box.
    # Find those cells and make copies to be placed on the other side.

    # Cells hanging off the right edge.
    add_x_right = (
        proj.field_data["px"] + 0.5 * proj.field_data["pdx"]
        > np.ceil(my_slice["box_width_fraction"]) * di_right_x
    )
    add_x_px = proj.field_data["px"][add_x_right]
    add_x_px -= np.ceil(my_slice["box_width_fraction"]) * di_right_x
    add_x_py = proj.field_data["py"][add_x_right]
    add_x_pdx = proj.field_data["pdx"][add_x_right]
    add_x_pdy = proj.field_data["pdy"][add_x_right]
    add_x_field = proj.field_data[proj_field][add_x_right]
    if weight_field is not None:
        add_x_weight_field = weight_proj.field_data[proj_weight_field][add_x_right]
    del add_x_right

    # Cells hanging off the left edge.
    add_x_left = proj.field_data["px"] - 0.5 * proj.field_data["pdx"] < di_left_x
    add2_x_px = proj.field_data["px"][add_x_left]
    add2_x_px += np.ceil(my_slice["box_width_fraction"]) * di_right_x
    add2_x_py = proj.field_data["py"][add_x_left]
    add2_x_pdx = proj.field_data["pdx"][add_x_left]
    add2_x_pdy = proj.field_data["pdy"][add_x_left]
    add2_x_field = proj.field_data[proj_field][add_x_left]
    if weight_field is not None:
        add2_x_weight_field = weight_proj.field_data[proj_weight_field][add_x_left]
    del add_x_left

    # Cells hanging off the top edge.
    add_y_right = (
        proj.field_data["py"] + 0.5 * proj.field_data["pdy"]
        > np.ceil(my_slice["box_width_fraction"]) * di_right_y
    )
    add_y_px = proj.field_data["px"][add_y_right]
    add_y_py = proj.field_data["py"][add_y_right]
    add_y_py -= np.ceil(my_slice["box_width_fraction"]) * di_right_y
    add_y_pdx = proj.field_data["pdx"][add_y_right]
    add_y_pdy = proj.field_data["pdy"][add_y_right]
    add_y_field = proj.field_data[proj_field][add_y_right]
    if weight_field is not None:
        add_y_weight_field = weight_proj.field_data[proj_weight_field][add_y_right]
    del add_y_right

    # Cells hanging off the bottom edge.
    add_y_left = proj.field_data["py"] - 0.5 * proj.field_data["pdy"] < di_left_y
    add2_y_px = proj.field_data["px"][add_y_left]
    add2_y_py = proj.field_data["py"][add_y_left]
    add2_y_py += np.ceil(my_slice["box_width_fraction"]) * di_right_y
    add2_y_pdx = proj.field_data["pdx"][add_y_left]
    add2_y_pdy = proj.field_data["pdy"][add_y_left]
    add2_y_field = proj.field_data[proj_field][add_y_left]
    if weight_field is not None:
        add2_y_weight_field = weight_proj.field_data[proj_weight_field][add_y_left]
    del add_y_left

    # Add the hanging cells back to the projection data.
    proj.field_data["px"] = uconcatenate(
        [proj.field_data["px"], add_x_px, add_y_px, add2_x_px, add2_y_px]
    )
    proj.field_data["py"] = uconcatenate(
        [proj.field_data["py"], add_x_py, add_y_py, add2_x_py, add2_y_py]
    )
    proj.field_data["pdx"] = uconcatenate(
        [proj.field_data["pdx"], add_x_pdx, add_y_pdx, add2_x_pdx, add2_y_pdx]
    )
    proj.field_data["pdy"] = uconcatenate(
        [proj.field_data["pdy"], add_x_pdy, add_y_pdy, add2_x_pdy, add2_y_pdy]
    )
    proj.field_data[proj_field] = uconcatenate(
        [
            proj.field_data[proj_field],
            add_x_field,
            add_y_field,
            add2_x_field,
            add2_y_field,
        ]
    )
    if weight_field is not None:
        weight_proj.field_data[proj_weight_field] = uconcatenate(
            [
                weight_proj.field_data[proj_weight_field],
                add_x_weight_field,
                add_y_weight_field,
                add2_x_weight_field,
                add2_y_weight_field,
            ]
        )

    # Delete original copies of hanging cells.
    del add_x_px, add_y_px, add2_x_px, add2_y_px
    del add_x_py, add_y_py, add2_x_py, add2_y_py
    del add_x_pdx, add_y_pdx, add2_x_pdx, add2_y_pdx
    del add_x_pdy, add_y_pdy, add2_x_pdy, add2_y_pdy
    del add_x_field, add_y_field, add2_x_field, add2_y_field
    if weight_field is not None:
        del (
            add_x_weight_field,
            add_y_weight_field,
            add2_x_weight_field,
            add2_y_weight_field,
        )

    # Tiles were made rounding up the width to the nearest integer.
    # Cut off the edges to get the specified width.
    # Cut in the x direction.
    cut_x = (
        proj.field_data["px"] - 0.5 * proj.field_data["pdx"]
        < di_right_x * my_slice["box_width_fraction"]
    )
    proj.field_data["px"] = proj.field_data["px"][cut_x]
    proj.field_data["py"] = proj.field_data["py"][cut_x]
    proj.field_data["pdx"] = proj.field_data["pdx"][cut_x]
    proj.field_data["pdy"] = proj.field_data["pdy"][cut_x]
    proj.field_data[proj_field] = proj.field_data[proj_field][cut_x]
    if weight_field is not None:
        weight_proj.field_data[proj_weight_field] = weight_proj.field_data[
            proj_weight_field
        ][cut_x]
    del cut_x

    # Cut in the y direction.
    cut_y = (
        proj.field_data["py"] - 0.5 * proj.field_data["pdy"]
        < di_right_y * my_slice["box_width_fraction"]
    )
    proj.field_data["px"] = proj.field_data["px"][cut_y]
    proj.field_data["py"] = proj.field_data["py"][cut_y]
    proj.field_data["pdx"] = proj.field_data["pdx"][cut_y]
    proj.field_data["pdy"] = proj.field_data["pdy"][cut_y]
    proj.field_data[proj_field] = proj.field_data[proj_field][cut_y]
    if weight_field is not None:
        weight_proj.field_data[proj_weight_field] = weight_proj.field_data[
            proj_weight_field
        ][cut_y]
    del cut_y

    # Create fixed resolution buffer to return back to the light cone object.
    # These buffers will be stacked together to make the light cone.
    frb = FixedResolutionBuffer(
        proj,
        (
            di_left_x,
            di_right_x * my_slice["box_width_fraction"],
            di_left_y,
            di_right_y * my_slice["box_width_fraction"],
        ),
        (pixels, pixels),
        antialias=False,
    )
    rdata = {"field": frb[proj_field]}

    if weight_field is not None:
        for f in ["px", "py", "pdx", "pdy"]:
            weight_proj.field_data[f] = proj.field_data[f]
        wfrb = FixedResolutionBuffer(
            weight_proj,
            (
                di_left_x,
                di_right_x * my_slice["box_width_fraction"],
                di_left_y,
                di_right_y * my_slice["box_width_fraction"],
            ),
            (pixels, pixels),
            antialias=False,
        )
        rdata["weight_field"] = wfrb[proj_weight_field]
        rdata["field"] *= rdata["weight_field"]

    return rdata



"""
light cone generator test



"""

# -----------------------------------------------------------------------------
# Copyright (c) 2017, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

import os
import shutil
import tempfile

import numpy as np

from yt.testing import assert_equal
from yt.units.yt_array import YTQuantity
from yt.utilities.answer_testing.framework import AnswerTestingTest
from yt.utilities.on_demand_imports import _h5py as h5py
from yt_astro_analysis.cosmological_observation.api import LightCone
from yt_astro_analysis.utilities.testing import requires_sim

ETC = "enzo_tiny_cosmology/32Mpc_32.enzo"
_funits = {
    "density": YTQuantity(1, "g/cm**3"),
    "temperature": YTQuantity(1, "K"),
    "length": YTQuantity(1, "cm"),
}


class LightConeProjectionTest(AnswerTestingTest):
    _type_name = "LightConeProjection"
    _attrs = ()

    def __init__(self, parameter_file, simulation_type, field, weight_field=None):
        self.parameter_file = parameter_file
        self.simulation_type = simulation_type
        self.ds = os.path.basename(self.parameter_file)
        self.field = field
        self.weight_field = weight_field

    @property
    def storage_name(self):
        return "_".join(
            (os.path.basename(self.parameter_file), self.field, str(self.weight_field))
        )

    def run(self):
        # Set up in a temp dir
        tmpdir = tempfile.mkdtemp()
        curdir = os.getcwd()
        os.chdir(tmpdir)

        lc = LightCone(
            self.parameter_file,
            self.simulation_type,
            0.0,
            0.1,
            observer_redshift=0.0,
            time_data=False,
        )
        lc.calculate_light_cone_solution(seed=123456789, filename="LC/solution.txt")
        lc.project_light_cone(
            (600.0, "arcmin"),
            (60.0, "arcsec"),
            self.field,
            weight_field=self.weight_field,
            save_stack=True,
        )

        dname = f"{self.field}_{self.weight_field}"
        fh = h5py.File("LC/LightCone.h5", mode="r")
        data = fh[dname][()]
        units = fh[dname].attrs["units"]
        if self.weight_field is None:
            punits = _funits[self.field] * _funits["length"]
        else:
            punits = (
                _funits[self.field] * _funits[self.weight_field] * _funits["length"]
            )
            wunits = fh["weight_field_%s" % self.weight_field].attrs["units"]
            pwunits = _funits[self.weight_field] * _funits["length"]
            assert wunits == str(pwunits.units)
        assert units == str(punits.units)
        fh.close()

        # clean up
        os.chdir(curdir)
        shutil.rmtree(tmpdir)

        mean = data.mean()
        mi = data[data.nonzero()].min()
        ma = data.max()
        return np.array([mean, mi, ma])

    def compare(self, new_result, old_result):
        assert_equal(new_result, old_result, verbose=True)


@requires_sim(ETC, "Enzo")
def test_light_cone_projection():
    yield LightConeProjectionTest(ETC, "Enzo", "density")
    yield LightConeProjectionTest(ETC, "Enzo", "temperature", weight_field="density")



"""
logging utilities



"""

# -----------------------------------------------------------------------------
# Copyright (c) yt Development Team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

from yt.utilities.logger import ytLogger


class quiet:
    def __init__(self, mylog=None, minlevel=40):
        if mylog is None:
            mylog = ytLogger
        self.mylog = mylog
        self.minlevel = minlevel
        self.level = mylog.level

    def __enter__(self):
        if self.level > 10 and self.level < self.minlevel:
            self.mylog.setLevel(40)

    def __exit__(self, *args):
        self.mylog.setLevel(self.level)



import warnings

import numpy as np

from yt.data_objects.data_containers import YTDataContainer
from yt.data_objects.static_output import Dataset
from yt.visualization.plot_modifications import PlotCallback


class HaloCatalogCallback(PlotCallback):
    """
    Plots circles at the locations of all the halos
    in a halo catalog with radii corresponding to the
    virial radius of each halo.

    Parameters
    ----------
    halo_catalog : :class:`~yt.data_objects.static_output.Dataset`,
        :class:`~yt.data_objects.data_containers.YTDataContainer`, or
        :class:`~yt_astro_analysis.halo_analysis.halo_catalog.halo_catalog.HaloCatalog`
        The object containing halos to be overplotted. This can
        be a HaloCatalog object, a loaded halo catalog dataset,
        or a data container from a halo catalog dataset.
    circle_args : list
        Contains the arguments controlling the
        appearance of the circles, supplied to the
        Matplotlib patch Circle.
    width : tuple
        The width over which to select halos to plot,
        useful when overplotting to a slice plot. Accepts
        a tuple in the form (1.0, 'Mpc').
    annotate_field : str
        A field contained in the
        halo catalog to add text to the plot near the halo.
        Example: annotate_field = 'particle_mass' will
        write the halo mass next to each halo.
    radius_field : str
        A field contained in the halo
        catalog to set the radius of the circle which will
        surround each halo. Default: 'virial_radius'.
    center_field_prefix : str
        Accepts a field prefix which will
        be used to find the fields containing the coordinates
        of the center of each halo. Ex: 'particle_position'
        will result in the fields 'particle_position_x' for x
        'particle_position_y' for y, and 'particle_position_z'
        for z. Default: 'particle_position'.
    text_args : dict
        Contains the arguments controlling the text
        appearance of the annotated field.
    factor : float
        A number the virial radius is multiplied by for
        plotting the circles. Ex: factor = 2.0 will plot
        circles with twice the radius of each halo virial radius.

    Examples
    --------

    >>> import yt
    >>> dds = yt.load("Enzo_64/DD0043/data0043")
    >>> hds = yt.load("rockstar_halos/halos_0.0.bin")
    >>> p = yt.ProjectionPlot(
    ...     dds, "x", ("gas", "density"), weight_field=("gas", "density")
    ... )
    >>> p.annotate_halos(hds)
    >>> p.save()

    >>> # plot a subset of all halos
    >>> import yt
    >>> dds = yt.load("Enzo_64/DD0043/data0043")
    >>> hds = yt.load("rockstar_halos/halos_0.0.bin")
    >>> # make a region half the width of the box
    >>> dregion = dds.box(
    ...     dds.domain_center - 0.25 * dds.domain_width,
    ...     dds.domain_center + 0.25 * dds.domain_width,
    ... )
    >>> hregion = hds.box(
    ...     hds.domain_center - 0.25 * hds.domain_width,
    ...     hds.domain_center + 0.25 * hds.domain_width,
    ... )
    >>> p = yt.ProjectionPlot(
    ...     dds,
    ...     "x",
    ...     ("gas", "density"),
    ...     weight_field=("gas", "density"),
    ...     data_source=dregion,
    ...     width=0.5,
    ... )
    >>> p.annotate_halos(hregion)
    >>> p.save()

    >>> # plot halos from a HaloCatalog
    >>> import yt
    >>> from yt.extensions.astro_analysis.halo_analysis import HaloCatalog
    >>> dds = yt.load("Enzo_64/DD0043/data0043")
    >>> hds = yt.load("rockstar_halos/halos_0.0.bin")
    >>> hc = HaloCatalog(data_ds=dds, halos_ds=hds)
    >>> p = yt.ProjectionPlot(
    ...     dds, "x", ("gas", "density"), weight_field=("gas", "density")
    ... )
    >>> p.annotate_halos(hc)
    >>> p.save()

    """

    _type_name = "halos"
    region = None
    _descriptor = None
    _supported_geometries = ("cartesian", "spectral_cube")

    def __init__(
        self,
        halo_catalog,
        circle_args=None,
        circle_kwargs=None,
        width=None,
        annotate_field=None,
        radius_field="virial_radius",
        center_field_prefix="particle_position",
        text_args=None,
        font_kwargs=None,
        factor=1.0,
    ):

        PlotCallback.__init__(self)
        def_circle_args = {"edgecolor": "white", "facecolor": "None"}
        def_text_args = {"color": "white"}

        if isinstance(halo_catalog, YTDataContainer):
            self.halo_data = halo_catalog
        elif isinstance(halo_catalog, Dataset):
            self.halo_data = halo_catalog.all_data()
        else:
            # assuming halo_catalog is a HaloCatalog instance
            # but do so with a EAFP pattern instead of LBYL to workaround
            # conflicting namespaces, see https://github.com/yt-project/yt_astro_analysis/issues/131
            try:
                if halo_catalog.data_source.ds == halo_catalog.halos_ds:
                    self.halo_data = halo_catalog.data_source
                else:
                    self.halo_data = halo_catalog.halos_ds.all_data()
            except AttributeError as exc:
                raise TypeError(
                    "halo_catalog argument must be a HaloCatalog object, "
                    "a dataset, or a data container. "
                    f"Received {halo_catalog} with type {type(halo_catalog)}"
                ) from exc

        self.width = width
        self.radius_field = radius_field
        self.center_field_prefix = center_field_prefix
        self.annotate_field = annotate_field
        if circle_kwargs is not None:
            circle_args = circle_kwargs
            warnings.warn(
                "The circle_kwargs keyword is deprecated.  Please "
                "use the circle_args keyword instead."
            )
        if font_kwargs is not None:
            text_args = font_kwargs
            warnings.warn(
                "The font_kwargs keyword is deprecated.  Please use "
                "the text_args keyword instead."
            )
        if circle_args is None:
            circle_args = def_circle_args
        self.circle_args = circle_args
        if text_args is None:
            text_args = def_text_args
        self.text_args = text_args
        self.factor = factor

    def __call__(self, plot):
        from matplotlib.patches import Circle

        data = plot.data
        x0, x1, y0, y1 = self._physical_bounds(plot)
        xx0, xx1, yy0, yy1 = self._plot_bounds(plot)

        halo_data = self.halo_data
        axis_names = plot.data.ds.coordinates.axis_name
        xax = plot.data.ds.coordinates.x_axis[data.axis]
        yax = plot.data.ds.coordinates.y_axis[data.axis]
        field_x = f"{self.center_field_prefix}_{axis_names[xax]}"
        field_y = f"{self.center_field_prefix}_{axis_names[yax]}"
        field_z = f"{self.center_field_prefix}_{axis_names[data.axis]}"

        # Set up scales for pixel size and original data
        pixel_scale = self._pixel_scale(plot)[0]
        units = plot.xlim[0].units

        # Convert halo positions to code units of the plotted data
        # and then to units of the plotted window
        px = halo_data[("all", field_x)][:].in_units(units)
        py = halo_data[("all", field_y)][:].in_units(units)

        xplotcenter = (plot.xlim[0] + plot.xlim[1]) / 2
        yplotcenter = (plot.ylim[0] + plot.ylim[1]) / 2

        xdomaincenter = plot.ds.domain_center[xax]
        ydomaincenter = plot.ds.domain_center[yax]

        xoffset = xplotcenter - xdomaincenter
        yoffset = yplotcenter - ydomaincenter

        xdw = plot.ds.domain_width[xax].to(units)
        ydw = plot.ds.domain_width[yax].to(units)

        modpx = np.mod(px - xoffset, xdw) + xoffset
        modpy = np.mod(py - yoffset, ydw) + yoffset

        px[modpx != px] = modpx[modpx != px]
        py[modpy != py] = modpy[modpy != py]

        px, py = self._convert_to_plot(plot, [px, py])

        # Convert halo radii to a radius in pixels
        radius = halo_data[("all", self.radius_field)][:].in_units(units)
        radius = np.array(radius * pixel_scale * self.factor)

        if self.width:
            pz = halo_data[("all", field_z)][:].in_units("code_length")
            c = data.center[data.axis]

            # I should catch an error here if width isn't in this form
            # but I dont really want to reimplement get_sanitized_width...
            width = data.ds.arr(self.width[0], self.width[1]).in_units("code_length")

            indices = np.where((pz > c - 0.5 * width) & (pz < c + 0.5 * width))

            px = px[indices]
            py = py[indices]
            radius = radius[indices]

        for x, y, r in zip(px, py, radius):
            plot._axes.add_artist(Circle(xy=(x, y), radius=r, **self.circle_args))

        plot._axes.set_xlim(xx0, xx1)
        plot._axes.set_ylim(yy0, yy1)

        if self.annotate_field:
            annotate_dat = halo_data[("all", self.annotate_field)]
            texts = [f"{float(dat):g}" for dat in annotate_dat]
            labels = []
            for pos_x, pos_y, t in zip(px, py, texts):
                labels.append(plot._axes.text(pos_x, pos_y, t, **self.text_args))

            # Set the font properties of text from this callback to be
            # consistent with other text labels in this figure
            self._set_font_properties(plot, labels, **self.text_args)



"""
API for RadMC3D Export code



"""

# -----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

from yt_astro_analysis.radmc3d_export.RadMC3DImageUtilities import read_radmc3d_image
from yt_astro_analysis.radmc3d_export.RadMC3DInterface import (
    RadMC3DSource,
    RadMC3DWriter,
)



"""
callbacks, filters, quantities, and recipes for AnalysisPipeline



"""

from yt.utilities.operator_registry import OperatorRegistry

callback_registry = OperatorRegistry()


def add_callback(name, function):
    callback_registry[name] = AnalysisCallback(function)


class AnalysisCallback:
    r"""
    An AnalysisCallback is a function that minimally takes in a target object
    and performs some analysis on it. This function may attach attributes
    to the target object, write out data, etc, but does not return anything.
    """

    def __init__(self, function, args=None, kwargs=None):
        self.function = function
        self.args = args
        if self.args is None:
            self.args = []
        self.kwargs = kwargs
        if self.kwargs is None:
            self.kwargs = {}

    def __call__(self, target):
        self.function(target, *self.args, **self.kwargs)
        return True


filter_registry = OperatorRegistry()


def add_filter(name, function):
    filter_registry[name] = AnalysisFilter(function)


class AnalysisFilter(AnalysisCallback):
    r"""
    An AnalysisFilter is a function that minimally takes a target object, performs
    some analysis, and returns either True or False. The return value determines
    whether analysis is continued.
    """

    def __init__(self, function, *args, **kwargs):
        AnalysisCallback.__init__(self, function, args, kwargs)

    def __call__(self, target):
        return self.function(target, *self.args, **self.kwargs)


quantity_registry = OperatorRegistry()


def add_quantity(name, function):
    quantity_registry[name] = AnalysisQuantity(function)


class AnalysisQuantity(AnalysisCallback):
    r"""
    An AnalysisQuantity is a function that takes minimally a target object,
    performs some analysis, and then returns a value.
    """

    def __init__(self, function, *args, **kwargs):
        AnalysisCallback.__init__(self, function, args, kwargs)

    def __call__(self, target):
        return self.function(target, *self.args, **self.kwargs)


recipe_registry = OperatorRegistry()


def add_recipe(name, function):
    recipe_registry[name] = AnalysisRecipe(function)


class AnalysisRecipe:
    r"""
    An AnalysisRecipe is a function that accepts an AnalysisPipeline and
    adds a series of callbacks, quantities, and filters.
    """

    def __init__(self, function, args=None, kwargs=None):
        self.function = function
        self.args = args
        if self.args is None:
            self.args = []
        self.kwargs = kwargs
        if self.kwargs is None:
            self.kwargs = {}

    def __call__(self, pipeline):
        return self.function(pipeline, *self.args, **self.kwargs)



"""
Halo Finding methods



"""

import numpy as np

from yt.data_objects.time_series import DatasetSeries
from yt.utilities.operator_registry import OperatorRegistry
from yt_astro_analysis.halo_analysis.halo_finding.halo_objects import (
    FOFHaloFinder,
    HOPHaloFinder,
)

finding_method_registry = OperatorRegistry()


def add_finding_method(name, function):
    finding_method_registry[name] = HaloFindingMethod(function)


class HaloFindingMethod:
    r"""
    A halo finding method is a callback that performs halo finding on a
    dataset and returns a new dataset that is the loaded halo finder output.
    """

    def __init__(self, function, args=None, kwargs=None):
        self.function = function
        self.args = args
        if self.args is None:
            self.args = []
        self.kwargs = kwargs
        if self.kwargs is None:
            self.kwargs = {}

    def __call__(self, hc):
        return self.function(hc, *self.args, **self.kwargs)


def _hop_method(hc, **finder_kwargs):
    r"""
    Run the Hop halo finding method.
    """

    ds = hc.data_ds
    if isinstance(ds, DatasetSeries):
        ts = ds
    else:
        ts = DatasetSeries([ds])

    for my_ds in ts:
        halo_list = HOPHaloFinder(my_ds, **finder_kwargs)
        _parse_halo_list(hc, halo_list)


add_finding_method("hop", _hop_method)


def _fof_method(hc, **finder_kwargs):
    r"""
    Run the FoF halo finding method.
    """

    ds = hc.data_ds
    if isinstance(ds, DatasetSeries):
        ts = ds
    else:
        ts = DatasetSeries([ds])

    for my_ds in ts:
        halo_list = FOFHaloFinder(my_ds, **finder_kwargs)
        _parse_halo_list(hc, halo_list)


add_finding_method("fof", _fof_method)


def _rockstar_method(hc, **finder_kwargs):
    r"""
    Run the Rockstar halo finding method.
    """

    from yt_astro_analysis.halo_analysis.halo_finding.rockstar import RockstarHaloFinder

    if "outbase" not in finder_kwargs:
        finder_kwargs["outbase"] = hc.output_basedir

    ds = hc.data_ds
    rh = RockstarHaloFinder(ds, **finder_kwargs)
    rh.run()


add_finding_method("rockstar", _rockstar_method)


def _parse_halo_list(hc, halo_list):
    r"""
    Save the halo list as a HaloCatalog.
    """

    ds = halo_list.ds

    # Set up fields that we want to pull from identified halos and their units
    fields = [
        "particle_identifier",
        "particle_mass",
        "virial_radius",
        "particle_position_x",
        "particle_position_y",
        "particle_position_z",
        "particle_velocity_x",
        "particle_velocity_y",
        "particle_velocity_z",
    ]
    units = ["", "Msun", "kpc"] + ["unitary"] * 3 + ["km/s"] * 3
    ud = dict(zip(fields, units))

    # Set up a dictionary based on those fields
    # with empty arrays where we will fill in their values
    num_halos = len(halo_list)
    halo_properties = {
        f: ds.arr(np.empty(num_halos), unit) for f, unit in zip(fields, units)
    }

    save_particles = getattr(halo_list, "save_particles", False)
    if save_particles:
        n_particles = np.zeros(num_halos, dtype=np.int32)

    # Iterate through the halos pulling out fields.
    for i, halo in enumerate(halo_list):
        halo_properties["particle_identifier"][i] = halo.id
        halo_properties["particle_mass"][i] = halo.virial_mass().to(ud["particle_mass"])
        halo_properties["virial_radius"][i] = halo.virial_radius().to(
            ud["virial_radius"]
        )

        com = halo.center_of_mass()
        halo_properties["particle_position_x"][i] = com[0].to(ud["particle_position_x"])
        halo_properties["particle_position_y"][i] = com[1].to(ud["particle_position_y"])
        halo_properties["particle_position_z"][i] = com[2].to(ud["particle_position_z"])

        bv = halo.bulk_velocity()
        halo_properties["particle_velocity_x"][i] = bv[0].to(ud["particle_velocity_x"])
        halo_properties["particle_velocity_y"][i] = bv[1].to(ud["particle_velocity_y"])
        halo_properties["particle_velocity_z"][i] = bv[2].to(ud["particle_velocity_z"])

        if save_particles:
            n_particles[i] = halo.indices.size

    if save_particles:
        member_ids = np.empty(n_particles.sum(), dtype=np.int64)
        if member_ids.size > 0:
            np.concatenate(
                [halo["particle_index"].astype(np.int64) for halo in halo_list],
                out=member_ids,
            )

        start = n_particles.cumsum() - n_particles
        halo_properties.update(
            {
                "ids": member_ids,
                "particle_number": n_particles,
                "particle_index_start": start,
            }
        )

    field_types = {field: "." for field in halo_properties if field != "ids"}
    if save_particles:
        field_types["ids"] = "particles"

    hc._save(ds=ds, data=halo_properties, field_types=field_types)



import os
import shutil
import sys
import tempfile

from yt.frontends.halo_catalog.data_structures import YTHaloCatalogDataset
from yt.frontends.rockstar.data_structures import RockstarDataset
from yt.loaders import load
from yt.utilities.answer_testing.framework import FieldValuesTest, requires_ds

_fields = (
    ("halos", "particle_position_x"),
    ("halos", "particle_position_y"),
    ("halos", "particle_position_z"),
    ("halos", "particle_mass"),
)

methods = {"fof": 2, "hop": 2, "rockstar": 3}
decimals = {"fof": 10, "hop": 10, "rockstar": 1}

etiny = "enzo_tiny_cosmology/DD0046/DD0046"


@requires_ds(etiny, big_data=True)
def test_halo_finders_single():
    from mpi4py import MPI

    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    filename = os.path.join(os.path.dirname(__file__), "run_halo_finder.py")
    for method in methods:
        comm = MPI.COMM_SELF.Spawn(
            sys.executable, args=[filename, method, tmpdir], maxprocs=methods[method]
        )
        comm.Disconnect()

        if method == "rockstar":
            hcfn = "halos_0.0.bin"
        else:
            hcfn = os.path.join("DD0046", "DD0046.0.h5")
        fn = os.path.join(tmpdir, "halo_catalogs", method, hcfn)

        ds = load(fn)
        if method == "rockstar":
            ds.parameters["format_revision"] = 2
            ds_type = RockstarDataset
        else:
            ds_type = YTHaloCatalogDataset
        assert isinstance(ds, ds_type)

        for field in _fields:
            my_test = FieldValuesTest(
                ds, field, particle_type=True, decimals=decimals[method]
            )
            my_test.suffix = method
            yield my_test

    os.chdir(curdir)
    shutil.rmtree(tmpdir)



"""
AnalysisPipeline class and member functions



"""

import os

from yt.funcs import ensure_dir
from yt_astro_analysis.halo_analysis.halo_catalog.analysis_operators import (
    callback_registry,
    filter_registry,
    quantity_registry,
    recipe_registry,
)


class AnalysisPipeline:
    def __init__(self, output_dir=None):
        self.actions = []
        self.quantities = []
        self.field_quantities = []
        if output_dir is None:
            output_dir = "."
        self.output_dir = output_dir

    def add_callback(self, callback, *args, **kwargs):
        callback = callback_registry.find(callback, *args, **kwargs)
        self.actions.append(("callback", callback))

    def add_quantity(self, key, *args, **kwargs):
        from_data_source = kwargs.pop("from_data_source", False)
        field_type = kwargs.pop("field_type", None)

        if not from_data_source:
            quantity = quantity_registry.find(key, *args, **kwargs)
        else:
            if field_type is None:
                quantity = key
            else:
                quantity = (field_type, key)
            self.field_quantities.append(quantity)

        self.quantities.append(key)
        self.actions.append(("quantity", (key, quantity)))

    def add_filter(self, halo_filter, *args, **kwargs):
        halo_filter = filter_registry.find(halo_filter, *args, **kwargs)
        self.actions.append(("filter", halo_filter))

    def add_recipe(self, recipe, *args, **kwargs):
        halo_recipe = recipe_registry.find(recipe, *args, **kwargs)
        halo_recipe(self)

    def _preprocess(self):
        "Create callback output directories."

        for action_type, action in self.actions:
            if action_type != "callback":
                continue
            my_output_dir = action.kwargs.get("output_dir")
            if my_output_dir is not None:
                new_output_dir = ensure_dir(
                    os.path.join(self.output_dir, my_output_dir)
                )
                action.kwargs["output_dir"] = new_output_dir

    def _process_target(self, target):
        target_filter = True
        for action_type, action in self.actions:
            if action_type == "callback":
                action(target)
            elif action_type == "filter":
                target_filter = action(target)
                if not target_filter:
                    break
            elif action_type == "quantity":
                key, quantity = action
                if callable(quantity):
                    target.quantities[key] = quantity(target)
                else:
                    target._set_field_value(key, quantity)
            else:
                raise RuntimeError("Action must be a callback, filter, or quantity.")

        return target_filter



import os
import sys

from mpi4py import MPI

import yt
from yt.data_objects.particle_filters import particle_filter
from yt_astro_analysis.halo_analysis import HaloCatalog

yt.enable_parallelism()

method = sys.argv[1]
data_dir = sys.argv[2]
comm = MPI.Comm.Get_parent()

methods = {
    "fof": {"ptype": "dark_matter"},
    "hop": {"ptype": "dark_matter"},
    "rockstar": {"num_readers": 1, "num_writers": 1, "particle_type": "dark_matter"},
}


@particle_filter("dark_matter", requires=["creation_time"])
def _dm_filter(pfilter, data):
    return data[pfilter.filtered_type, "creation_time"] <= 0.0


ds = yt.load("enzo_tiny_cosmology/DD0046/DD0046")
ds.add_particle_filter("dark_matter")

output_dir = os.path.join(data_dir, "halo_catalogs", method)
hc = HaloCatalog(
    data_ds=ds,
    output_dir=output_dir,
    finder_method=method,
    finder_kwargs=methods[method],
)
hc.create()

comm.Disconnect()



"""
API for yt astro analysis.



"""

# -----------------------------------------------------------------------------
# Copyright (c) 2017, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

__version__ = "1.1.1"

from yt_astro_analysis import (
    cosmological_observation,
    halo_analysis,
    ppv_cube,
    radmc3d_export,
)



"""
HaloCatalog class and member functions



"""

import functools
import os

import numpy as np
from unyt import unyt_array

from yt.data_objects.time_series import DatasetSeries
from yt.frontends.ytdata.utilities import save_as_dataset
from yt.funcs import ensure_dir, mylog
from yt.utilities.parallel_tools.parallel_analysis_interface import (
    ParallelAnalysisInterface,
    parallel_blocking_call,
    parallel_objects,
)
from yt_astro_analysis.halo_analysis.halo_catalog.analysis_pipeline import (
    AnalysisPipeline,
)
from yt_astro_analysis.halo_analysis.halo_catalog.halo_finding_methods import (
    finding_method_registry,
)
from yt_astro_analysis.halo_analysis.halo_catalog.halo_object import Halo
from yt_astro_analysis.utilities.logging import quiet

_default_fields = (
    ["particle_identifier", "particle_mass", "virial_radius"]
    + ["particle_position_" + ax for ax in "xyz"]
    + ["particle_velocity_" + ax for ax in "xyz"]
)


class HaloCatalog(ParallelAnalysisInterface):
    r"""Create a HaloCatalog: an object that allows for the creation and association
    of data with a set of halo objects.

    A HaloCatalog object pairs a simulation dataset and the output from a halo finder,
    allowing the user to perform analysis on each of the halos found by the halo finder.
    Analysis is performed by providing callbacks: functions that accept a Halo object
    and perform independent analysis, return a quantity to be associated with the halo,
    or return True or False whether a halo meets various criteria.  The resulting set of
    quantities associated with each halo is then written out to disk at a "halo catalog."
    This halo catalog can then be loaded in with yt as any other simulation dataset.

    Parameters
    ----------
    halos_ds : str
        Dataset created by a halo finder.  If None, a halo finder should be
        provided with the finder_method keyword.
    data_ds : str
        Dataset created by a simulation.
    data_source : data container
        Data container associated with either the halos_ds to use for analysis.
        This can be used to restrict analysis to a subset of the full catalog.
        By default, the entire catalog will be analyzed.
    halo_field_type : str
        The field type for halos. This can be used to specify a certain type of halo
        in a dataset that contains multiple types.
        Default: "all"
    finder_method : str
        Halo finder to be used if no halos_ds is given.
    output_dir : str
        The top level directory into which analysis output will be written.
        Default: "halo_catalogs"
    finder_kwargs : dict
        Arguments to pass to the halo finder if finder_method is given.

    Examples
    --------

    >>> # create profiles or overdensity vs. radius for each halo and save to disk
    >>> import yt
    >>> from yt.extensions.astro_analysis.halo_analysis import HaloCatalog
    >>> data_ds = yt.load("DD0064/DD0064")
    >>> halos_ds = yt.load("rockstar_halos/halos_64.0.bin",
    ...                    output_dir="halo_catalogs/catalog_0064")
    >>> hc = HaloCatalog(data_ds=data_ds, halos_ds=halos_ds)
    >>> # filter out halos with mass < 1e13 Msun
    >>> hc.add_filter("quantity_value", "particle_mass", ">", 1e13, "Msun")
    >>> # create a sphere object with radius of 2 times the virial_radius field
    >>> hc.add_callback("sphere", factor=2.0, radius_field="virial_radius")
    >>> # make radial profiles
    >>> hc.add_callback("profile", "radius", [("gas", "overdensity")],
    ...                 weight_field="cell_volume", accumulation=True)
    >>> # save the profiles to disk
    >>> hc.add_callback("save_profiles", output_dir="profiles")
    >>> # create the catalog
    >>> hc.create()

    >>> # load in the saved halo catalog and all the profile data
    >>> halos_ds = yt.load("halo_catalogs/catalog_0064/catalog_0064.0.h5")
    >>> hc = HaloCatalog(halos_ds=halos_ds,
                         output_dir="halo_catalogs/catalog_0064")
    >>> hc.add_callback("load_profiles", output_dir="profiles")
    >>> hc.load()

    See Also
    --------
    add_callback, add_filter, add_quantity, add_recipe

    """

    _id_field = "particle_identifier"

    def __init__(
        self,
        halos_ds=None,
        data_ds=None,
        data_source=None,
        halo_field_type="all",
        finder_method=None,
        finder_kwargs=None,
        output_dir=None,
    ):

        super().__init__()

        self.halos_ds = halos_ds
        self.data_ds = data_ds
        self.halo_field_type = halo_field_type

        if halos_ds is None:
            if data_ds is None:
                raise RuntimeError("Must specify a halos_ds, data_ds, or both.")
            if finder_method is None:
                raise RuntimeError("Must specify a halos_ds or a finder_method.")

        if data_source is None and halos_ds is not None:
            data_source = halos_ds.all_data()
        self.data_source = data_source

        if output_dir is None:
            if finder_method == "rockstar":
                output_dir = finder_kwargs.get("outbase", "rockstar_halos")
            else:
                output_dir = "halo_catalogs"

        self.output_basedir = ensure_dir(output_dir)
        self.pipeline = AnalysisPipeline(output_dir=self.output_dir)
        self.quantities = self.pipeline.quantities

        self.finder_method_name = finder_method
        if finder_kwargs is None:
            finder_kwargs = {}
        if finder_method is not None:
            finder_method = finding_method_registry.find(finder_method, **finder_kwargs)
        self.finder_method = finder_method

        self._add_default_quantities()

    def _add_default_quantities(self):
        if self.halos_ds is None:
            return

        field_type = self.halo_field_type
        for field in _default_fields:
            field_name = (field_type, field)
            if field_name not in self.halos_ds.derived_field_list:
                mylog.warning(
                    "Halo dataset %s has no field %s.", self.halos_ds, field_name
                )
                continue
            self.add_quantity(field, from_data_source=True, field_type=field_type)

    _source_ds = None

    @property
    def source_ds(self):
        if self._source_ds is not None:
            return self._source_ds
        if self.data_source is not None:
            return self.data_source.ds
        return None

    @property
    def output_basename(self):
        ds = self.data_ds
        if isinstance(ds, DatasetSeries):
            ds = None

        if ds is None:
            ds = self.source_ds
        if ds is None:
            basename = "halo_catalog"
        else:
            basename = ds.basename

        if "." in basename:
            basename = basename[: basename.find(".")]
        return basename

    @property
    def output_dir(self):
        return os.path.join(self.output_basedir, self.output_basename)

    def _yield_halos(self, njobs="auto", dynamic=False):

        my_size = self.comm.size

        if njobs == "auto":
            # use task queue if odd number of cores more than 2
            my_dynamic = my_size > 2 and my_size % 2
            my_njobs = -1
        else:
            my_dynamic = dynamic
            my_njobs = njobs

        for chunk in self.data_source.chunks([], "io"):

            if self.comm.rank == 0:
                chunk.get_data(self.pipeline.field_quantities)

            if my_size > 1:
                fdata = self.comm.comm.bcast(chunk.field_data, root=0)
                chunk.field_data.update(fdata)

            target_indices = range(chunk[self.halo_field_type, self._id_field].size)
            my_indices = parallel_objects(
                target_indices, njobs=my_njobs, dynamic=my_dynamic
            )

            for my_index in my_indices:
                my_halo = Halo(self, chunk, my_index)
                yield my_halo

    @parallel_blocking_call
    def _run(self, save_halos, save_catalog, njobs="auto", dynamic=False):
        """
        Run analysis pipeline on all halos.

        This is used by both load and create.
        """

        # Find halos.
        if self.halos_ds is None:
            self.finder_method(self)
            return

        self.pipeline._preprocess()

        self.catalog = []
        if save_halos:
            self.halo_list = []

        for my_halo in self._yield_halos(njobs=njobs, dynamic=dynamic):
            rval = self.pipeline._process_target(my_halo)

            if rval:
                for quantity in my_halo.quantities.values():
                    if hasattr(quantity, "units"):
                        quantity.convert_to_base()
                self.catalog.append(my_halo.quantities)

            if save_halos and rval:
                self.halo_list.append(my_halo)
            else:
                del my_halo

        if save_catalog:
            self._save()

    def _save(self, ds=None, data=None, extra_attrs=None, field_types=None):
        "Save new halo catalog."

        if ds is None:
            ds = self.source_ds
        else:
            self._source_ds = ds

        data_dir = ensure_dir(self.output_dir)
        filename = os.path.join(data_dir, f"{self.output_basename}.{self.comm.rank}.h5")

        if data is None:
            n_halos = len(self.catalog)
            data = {}
            if n_halos > 0:
                for key in self.quantities:

                    if hasattr(self.catalog[0][key], "units"):
                        registry = self.catalog[0][key].units.registry
                        my_arr = functools.partial(unyt_array, registry=registry)
                    else:
                        my_arr = np.array

                    data[key] = my_arr([halo[key] for halo in self.catalog])
        else:
            n_halos = data[self._id_field].size

        mylog.info("Saving %d halos: %s.", n_halos, filename)

        if field_types is None:
            field_types = {key: "." for key in self.quantities}

        if extra_attrs is None:
            extra_attrs = {}
        extra_attrs_d = {"data_type": "halo_catalog", "num_halos": n_halos}
        extra_attrs_d.update(extra_attrs)

        with quiet():
            save_as_dataset(
                ds, filename, data, field_types=field_types, extra_attrs=extra_attrs_d
            )

    def create(self, save_halos=False, save_output=True, njobs="auto", dynamic=False):
        r"""
        Create the halo catalog given the callbacks, quantities, and filters that
        have been provided.

        This is a wrapper around the main _run function with default arguments tuned
        for halo catalog creation.  By default, halo objects are not saved but the
        halo catalog is written, opposite to the behavior of the load function.

        Parameters
        ----------
        save_halos : bool
            If True, a list of all Halo objects is retained under the "halo_list"
            attribute.  If False, only the compiles quantities are saved under the
            "catalog" attribute.
            Default: False
        save_output : bool
            If True, save the final catalog to disk.
            Default: True
        njobs : int
            The number of jobs over which to divide halo analysis. If set to "auto",
            use a task queue if total number of processors is an odd number and
            divide jobs evenly if an even number.
            Default: "auto"
        dynamic : int
            If False, halo analysis is divided evenly between all available processors.
            If True, parallelism is performed via a task queue. If njobs is set to
            "auto", behavior is controlled in the way described above.
            Default: False

        See Also
        --------
        load

        """

        self._run(save_halos, save_output, njobs=njobs, dynamic=dynamic)

    def load(self, njobs="auto", dynamic=False):
        r"""
        Load a previously created halo catalog.

        This is a wrapper around the main _run function with default arguments tuned
        for reloading halo catalogs and associated data.  By default, halo objects are
        saved and the halo catalog is not written, opposite to the behavior of the
        create function.

        Parameters
        ----------
        njobs : int
            The number of jobs over which to divide halo analysis. If set to "auto",
            use a task queue if total number of processors is an odd number and
            divide jobs evenly if an even number.
            Default: "auto"
        dynamic : int
            If False, halo analysis is divided evenly between all available processors.
            If True, parallelism is performed via a task queue. If njobs is set to
            "auto", behavior is controlled in the way described above.
            Default: False

        See Also
        --------
        create

        """

        self._run(True, False, njobs=njobs, dynamic=dynamic)

    def add_callback(self, callback, *args, **kwargs):
        r"""
        Add a callback to the halo catalog action list.

        A callback is a function that accepts and operates on a Halo object and
        does not return anything.  Callbacks must exist within the callback_registry.
        Give additional args and kwargs to be passed to the callback here.

        Parameters
        ----------
        callback : string
            The name of the callback.

        Examples
        --------

        >>> # Here, a callback is defined and added to the registry.
        >>> def _say_something(halo, message):
        ...     my_id = halo.quantities['particle_identifier']
        ...     print "Halo %d: here is a message - %s." % (my_id, message)
        >>> add_callback("hello_world", _say_something)

        >>> # Now this callback is accessible to the HaloCatalog object
        >>> hc.add_callback("hello_world", "this is my message")

        """

        self.pipeline.add_callback(callback, *args, **kwargs)

    def add_quantity(self, key, *args, **kwargs):
        r"""
        Add a quantity to the halo catalog action list.

        A quantity is a function that accepts a Halo object and return a value or
        values.  These values are stored in a "quantities" dictionary associated
        with the Halo object.  Quantities must exist within the quantity_registry.
        Give additional args and kwargs to be passed to the quantity function here.

        Parameters
        ----------
        key : string
            The name of the callback.
        field_type : string
            If not None, the quantity is the value of the field provided by the
            key parameter, taken from the halo finder dataset.  This is the way
            one pulls values for the halo from the halo dataset.
            Default : None

        Examples
        --------

        >>> # pull the virial radius from the halo finder dataset
        >>> hc.add_quantity("virial_radius", field_type="halos")

        >>> # define a custom quantity and add it to the register
        >>> def _mass_squared(halo):
        ...     # assume some entry "particle_mass" exists in the quantities dict
        ...     return halo.quantities["particle_mass"]**2
        >>> add_quantity("mass_squared", _mass_squared)

        >>> # add it to the halo catalog action list
        >>> hc.add_quantity("mass_squared")

        """

        self.pipeline.add_quantity(key, *args, **kwargs)

    def add_filter(self, halo_filter, *args, **kwargs):
        r"""
        Add a filter to the halo catalog action list.

        A filter is a function that accepts a Halo object and returns either True
        or False.  If True, any additional actions added to the list are carried out
        and the results are added to the final halo catalog.  If False, any further
        actions are skipped and the halo will be omitted from the final catalog.
        Filters must exist within the filter_registry.  Give additional args and kwargs
        to be passed to the filter function here.

        Parameters
        ----------
        halo_filter : string
            The name of the filter.

        Examples
        --------

        >>> # define a filter and add it to the register.
        >>> def _my_filter(halo, mass_value):
        ...     return halo.quantities["particle_mass"] > unyt_quantity(mass_value, "Msun")
        >>> # add it to the register
        >>> add_filter("mass_filter", _my_filter)

        >>> # add the filter to the halo catalog actions
        >>> hc.add_filter("mass_value", 1e12)

        """

        self.pipeline.add_filter(halo_filter, *args, **kwargs)

    def add_recipe(self, recipe, *args, **kwargs):
        r"""
        Add a recipe to the halo catalog action list.

        A recipe is an operation consisting of a series of callbacks, quantities,
        and/or filters called in succession.  Recipes can be used to store a more
        complex series of analysis tasks as a single entity.

        Currently, the available recipe is ``calculate_virial_quantities``.

        Parameters
        ----------

        halo_recipe : string
            The name of the recipe.

        Examples
        --------

        >>> import yt
        >>> from yt.extensions.astro_analysis.halo_analysis import HaloCatalog
        >>>
        >>> data_ds = yt.load('Enzo_64/RD0006/RedshiftOutput0006')
        >>> halos_ds = yt.load('rockstar_halos/halos_0.0.bin')
        >>> hc = HaloCatalog(data_ds=data_ds, halos_ds=halos_ds)
        >>>
        >>> # Filter out less massive halos
        >>> hc.add_filter("quantity_value", "particle_mass", ">", 1e14, "Msun")
        >>>
        >>> # Calculate virial radii
        >>> hc.add_recipe("calculate_virial_quantities", ["radius", "matter_mass"])
        >>>
        >>> hc.create()

        """

        self.pipeline.add_recipe(recipe, *args, **kwargs)



"""
Operations to run Rockstar



"""

# -----------------------------------------------------------------------------
# Copyright (c) yt Development Team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

from unyt import unyt_quantity

from yt.config import ytcfg
from yt.data_objects.static_output import Dataset
from yt.data_objects.time_series import DatasetSeries
from yt.funcs import is_root, mylog
from yt.utilities.parallel_tools.parallel_analysis_interface import (
    ParallelAnalysisInterface,
    ProcessorPool,
)

try:
    from yt_astro_analysis.halo_analysis.halo_finding.rockstar import rockstar_interface
except ImportError:
    mylog.warning(
        "Cannot import the rockstar interface.  Rockstar will not run.\n"
        "If you need Rockstar, see the installation instructions at "
        "http://yt-astro-analysis.readthedocs.io/."
    )
    rockstar_interface = None

import os
import socket
import time

import numpy as np


class InlineRunner(ParallelAnalysisInterface):
    def __init__(self):
        # If this is being run inline, num_readers == comm.size, always.
        psize = ytcfg.get("yt", "internals", "global_parallel_size")
        self.num_readers = psize
        # No choice for you, everyone's a writer too!
        self.num_writers = psize

    def run(self, handler, pool):
        # If inline, we use forks.
        server_pid = 0
        # Start a server on only one machine/fork.
        if pool.comm.rank == 0:
            server_pid = os.fork()
            if server_pid == 0:
                handler.start_server()
                os._exit(0)
        # Start writers on all.
        writer_pid = 0
        time.sleep(0.05 + pool.comm.rank / 10.0)
        writer_pid = os.fork()
        if writer_pid == 0:
            handler.start_writer()
            os._exit(0)
        # Everyone's a reader!
        time.sleep(0.05 + pool.comm.rank / 10.0)
        handler.start_reader()
        # Make sure the forks are done, which they should be.
        if writer_pid != 0:
            os.waitpid(writer_pid, 0)
        if server_pid != 0:
            os.waitpid(server_pid, 0)

    def setup_pool(self):
        pool = ProcessorPool()
        # Everyone is a reader, and when we're inline, that's all that matters.
        readers = np.arange(ytcfg.get("yt", "internals", "global_parallel_size"))
        pool.add_workgroup(ranks=readers, name="readers")
        return pool, pool.workgroups[0]


class StandardRunner(ParallelAnalysisInterface):
    def __init__(self, num_readers, num_writers):
        self.num_readers = num_readers
        psize = ytcfg.get("yt", "internals", "global_parallel_size")
        if num_writers is None:
            self.num_writers = psize - num_readers - 1
        else:
            self.num_writers = min(num_writers, psize)
        if self.num_readers + self.num_writers + 1 != psize:
            raise RuntimeError(
                "The number of MPI processes (%i) does not equal the "
                "number of readers (%i) plus the number of writers "
                "(%i) plus 1 server" % (psize, self.num_readers, self.num_writers)
            )

    def run(self, handler, wg):
        # Not inline so we just launch them directly from our MPI threads.
        if wg.name == "server":
            handler.start_server()
        if wg.name == "readers":
            time.sleep(0.05)
            handler.start_reader()
        if wg.name == "writers":
            time.sleep(0.1)
            handler.start_writer()

    def setup_pool(self):
        pool = ProcessorPool()
        pool, workgroup = ProcessorPool.from_sizes(
            [
                (1, "server"),
                (self.num_readers, "readers"),
                (self.num_writers, "writers"),
            ]
        )
        return pool, workgroup


class RockstarHaloFinder(ParallelAnalysisInterface):
    r"""Spawns the Rockstar Halo finder, distributes particles and finds halos.

    Rockstar has three main processes: reader, writer, and the
    server which coordinates reader/writer processes.

    Parameters
    ----------
    ts : ~yt.data_objects.time_series.DatasetSeries, ~yt.data_objects.static_output.Dataset
        The dataset or datsets on which halo finding will be run. If you intend
        to make a merger tree later, you must run Rockstar using a DatasetSeries
        containing all the snapshot to be included.
    num_readers : int
        The number of reader can be increased from the default
        of 1 in the event that a single snapshot is split among
        many files. This can help in cases where performance is
        IO-limited. Default is 1. If run inline, it is
        equal to the number of MPI threads.
    num_writers : int
        The number of writers determines the number of processing threads
        as well as the number of threads writing output data.
        The default is set to comm.size-num_readers-1. If run inline,
        the default is equal to the number of MPI threads.
    outbase : str
        This is where the out*list files that Rockstar makes should be
        placed. Default is 'rockstar_halos'.
    particle_type : str
        This is the "particle type" that can be found in the data.  This can be
        a filtered particle or an inherent type.
    mass_field : optional, str
        The field to be used for the particle masses. The sampled field will be
        (<particle_type>, <mass_field>). This can be used to provide alternative
        particle masses for halo finding.
        Default: "particle_mass"
    star_types : str list/array
        The types (as returned by data((particle_type, particle_type)) to be
        recognized as star particles.
    force_res : float
        This parameter specifies the force resolution that Rockstar uses
        in units of Mpccm/h (comoving Mpc/h).
        If no value is provided, this parameter is automatically set to
        the width of the smallest grid element in the simulation from the
        last data snapshot (i.e. the one where time has evolved the
        longest) in the time series:
        ``ds_last.index.get_smallest_dx().to("Mpccm/h")``.
    initial_metric_scaling : float
        The position element of the fof distance metric is divided by this
        parameter, set to 1 by default. If the initial_metric_scaling=0.1 the
        position element will have 10 times more weight than the velocity element,
        biasing the metric towards position information more so than velocity
        information. That was found to be needed for hydro-ART simulations
        with 10's of parsecs resolution.
        Default: 1.0.
    non_dm_metric_scaling : float
        The metric scaling to be used for non-dm particles. The effect of
        this parameter is currently unknown.
        Default: 10.
    suppress_galaxies : int
        Wether to include non-dm halos (i.e. galaxies) in the catalogs. The effect
        of this parameter is currently unknown.
        Default: 1.
    total_particles : int
        If supplied, this is a pre-calculated total number of particles present
        in the simulation. For example, this is useful when analyzing a series
        of snapshots where the number of dark matter particles should not
        change and this will save some disk access time. If left unspecified,
        it will be calculated automatically. Default: ``None``.
    particle_mass : optional, None, float, tuple, or :class:`~unyt.array.unyt_quantity`
        If supplied, this mass will be used to calculate the average particle
        spacing used in the friend-of-friends algorithm. The particle spacing
        will be (particle_mass / omega_matter * rho_cr)^1/3. If None, the mass
        is set as the minimum of all particles to be read. If a float, units
        are assumed to be in Msun/h. If a tuple, the format is assume to be
        (<value>, <units>). If a :class:`~unyt.array.unyt_quantity`, it must
        be convertible to units of Msun/h. To modify the masses of particles
        used for halo finding, see the mass_field keyword.
        Default: ``None``.
    restart : optional, bool
        Set to True to have rockstar restart from the first uncompleted
        snapshot. If False, rockstar will start at the first snapshot in the
        simulation.
        Default: False

    Returns
    -------
    None

    Examples
    --------

    To use the script below you must run it using MPI:
    `mpirun -np 4 python run_rockstar.py`

    >>> import yt
    >>> yt.enable_parallelism()
    >>> from yt.extensions.astro_analysis.halo_analysis import HaloCatalog
    >>> data_ds = yt.load('Enzo_64/RD0006/RedshiftOutput0006')
    >>> hc = HaloCatalog(data_ds=data_ds, finder_method='rockstar',
    ...                  finder_kwargs={"num_readers": 1, "num_writers": 2})
    >>> hc.create()

    """

    def __init__(
        self,
        ts,
        num_readers=1,
        num_writers=None,
        outbase="rockstar_halos",
        particle_type="all",
        mass_field="particle_mass",
        star_types=None,
        force_res=None,
        initial_metric_scaling=1.0,
        non_dm_metric_scaling=10.0,
        suppress_galaxies=1,
        total_particles=None,
        dm_only=False,
        particle_mass=None,
        min_halo_size=25,
        restart=False,
    ):

        if is_root():
            mylog.info("The citation for the Rockstar halo finder can be found at")
            mylog.info("http://adsabs.harvard.edu/abs/2013ApJ...762..109B")
        ParallelAnalysisInterface.__init__(self)
        # Decide how we're working.
        if ytcfg.get("yt", "inline"):
            self.runner = InlineRunner()
        else:
            self.runner = StandardRunner(num_readers, num_writers)
        self.restart = restart
        self.num_readers = self.runner.num_readers
        self.num_writers = self.runner.num_writers
        mylog.info(
            "Rockstar is using %d readers and %d writers",
            self.num_readers,
            self.num_writers,
        )
        # Note that Rockstar does not support subvolumes.
        # We assume that all of the snapshots in the time series
        # use the same domain info as the first snapshots.
        if not isinstance(ts, DatasetSeries):
            ts = DatasetSeries([ts])
        self.ts = ts
        self.particle_type = particle_type
        if star_types is None:
            star_types = []
        self.star_types = star_types
        self.outbase = outbase
        self.min_halo_size = min_halo_size
        if force_res is None:
            tds = ts[-1]  # Cache a reference
            self.force_res = tds.index.get_smallest_dx().to("Mpccm/h")
            # We have to delete now to wipe the index
            del tds
        else:
            self.force_res = force_res
        self.initial_metric_scaling = initial_metric_scaling
        self.non_dm_metric_scaling = non_dm_metric_scaling
        self.suppress_galaxies = suppress_galaxies
        self.total_particles = total_particles
        self.dm_only = dm_only
        self.particle_mass = particle_mass
        self.mass_field = mass_field
        # Setup pool and workgroups.
        self.pool, self.workgroup = self.runner.setup_pool()
        p = self._setup_parameters(ts)
        params = self.comm.mpi_bcast(p, root=self.pool["readers"].ranks[0])
        self.__dict__.update(params)
        self.handler = rockstar_interface.RockstarInterface(self.ts)

    def _setup_parameters(self, ts):
        if self.workgroup.name != "readers":
            return None
        tds = ts[0]
        tds.index
        ptype = self.particle_type
        if ptype not in tds.particle_types and ptype != "all":
            has_particle_filter = tds.add_particle_filter(ptype)
            if not has_particle_filter:
                raise RuntimeError("Particle type (filter) %s not found." % (ptype))

        dd = tds.all_data()
        # Get DM particle mass.
        particle_mass = self.particle_mass
        if particle_mass is None:
            pmass_min, pmass_max = dd.quantities.extrema(
                (ptype, self.mass_field), non_zero=True
            )
            particle_mass = pmass_min
        elif isinstance(particle_mass, (tuple, list)) and len(particle_mass) == 2:
            particle_mass = tds.quan(*particle_mass)
        elif not isinstance(particle_mass, unyt_quantity):
            particle_mass = tds.quan(particle_mass, "Msun / h")
        particle_mass.convert_to_units("Msun / h")

        p = {}
        if self.total_particles is None:
            # Get total_particles in parallel.
            tp = dd.quantities.total_quantity((ptype, "particle_ones"))
            p["total_particles"] = int(tp)
            mylog.info("Total Particle Count: %d.", int(tp))
        p["left_edge"] = tds.domain_left_edge.in_units("Mpccm/h")
        p["right_edge"] = tds.domain_right_edge.in_units("Mpccm/h")
        p["center"] = tds.domain_center.in_units("Mpccm/h")
        p["particle_mass"] = self.particle_mass = particle_mass
        del tds
        return p

    def __del__(self):
        try:
            self.pool.free_all()
        except AttributeError:
            # This really only acts to cut down on the misleading
            # error messages when/if this class is called incorrectly
            # or some other error happens and self.pool hasn't been created
            # already.
            pass

    def _get_hosts(self):
        if self.comm.rank == 0 or self.comm.size == 1:

            # Temporary mac hostname fix
            try:
                server_address = socket.gethostname()
                socket.gethostbyname(server_address)
            except socket.gaierror:
                server_address = "localhost"

            sock = socket.socket()
            sock.bind(("", 0))
            port = sock.getsockname()[-1]
            del sock
        else:
            server_address, port = None, None
        self.server_address, self.port = self.comm.mpi_bcast((server_address, port))
        self.server_address = bytearray(str(self.server_address), "utf-8")
        self.port = bytearray(str(self.port), "utf-8")

    def run(self, block_ratio=1, callbacks=None, restart=False):
        """ """
        if block_ratio != 1:
            raise NotImplementedError
        self._get_hosts()
        # Find restart output number
        num_outputs = len(self.ts)
        restart = restart or self.restart
        if restart:
            restart_file = os.path.join(self.outbase, "restart.cfg")
            if not os.path.exists(restart_file):
                raise RuntimeError("Restart file %s not found" % (restart_file))
            with open(restart_file) as restart_fh:
                for par in restart_fh:
                    if par.startswith("RESTART_SNAP"):
                        restart_num = int(par.split("=")[1])
                    if par.startswith("NUM_WRITERS"):
                        num_writers = int(par.split("=")[1])
            if num_writers != self.num_writers:
                raise RuntimeError(
                    "Number of writers in restart has changed from the original "
                    "run (OLD = %d, NEW = %d).  To avoid problems in the "
                    "restart, choose the same number of writers."
                    % (num_writers, self.num_writers)
                )
            # Remove the datasets that were already analyzed
            self.ts._pre_outputs = self.ts._pre_outputs[restart_num:]
        else:
            restart_num = 0
        outbase = bytearray(self.outbase, "utf-8")
        self.handler.setup_rockstar(
            self.server_address,
            self.port,
            num_outputs,
            np.int64(self.total_particles),
            self.particle_type,
            self.mass_field,
            star_types=self.star_types,
            particle_mass=self.particle_mass,
            parallel=self.comm.size > 1,
            num_readers=self.num_readers,
            num_writers=self.num_writers,
            writing_port=-1,
            block_ratio=block_ratio,
            outbase=outbase,
            force_res=self.force_res,
            initial_metric_scaling=self.initial_metric_scaling,
            non_dm_metric_scaling=self.non_dm_metric_scaling,
            suppress_galaxies=self.suppress_galaxies,
            callbacks=callbacks,
            restart_num=restart_num,
            min_halo_size=self.min_halo_size,
        )
        # Make the directory to store the halo lists in.
        if not self.outbase:
            self.outbase = os.getcwd()
        if self.comm.rank == 0 and not restart:
            if not os.path.exists(self.outbase):
                os.makedirs(self.outbase)
            # Make a record of which dataset corresponds to which set of
            # output files because it will be easy to lose this connection.
            fp = open(os.path.join(self.outbase, "datasets.txt"), "w")
            fp.write("# dsname\tindex\n")
            for i, ds in enumerate(self.ts.outputs):
                if isinstance(ds, Dataset):
                    fn = ds.parameter_filename
                else:
                    fn = ds
                dsloc = os.path.join(os.path.relpath(fn))
                line = f"{dsloc}\t{i}\n"
                fp.write(line)
            fp.close()
        # This barrier makes sure the directory exists before it might be used.
        self.comm.barrier()
        if self.comm.size == 1:
            self.handler.call_rockstar()
        else:
            # And run it!
            self.runner.run(self.handler, self.workgroup)
        self.comm.barrier()
        self.pool.free_all()



"""
Unit test the PPVCube analysis module.
"""

# -----------------------------------------------------------------------------
# Copyright (c) 2013-2017, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

import numpy as np

import yt.units as u
from yt.loaders import load_uniform_grid
from yt.testing import assert_allclose_units
from yt.utilities.physical_constants import clight, kboltz, mh
from yt_astro_analysis.ppv_cube.api import PPVCube


def setup():
    """Test specific setup."""
    from yt.config import ytcfg

    ytcfg["yt", "__withintesting"] = "True"


def test_ppv():

    np.random.seed(seed=0x4D3D3D3)

    dims = (8, 8, 128)
    v_shift = 1.0e7 * u.cm / u.s
    sigma_v = 2.0e7 * u.cm / u.s
    T_0 = 1.0e8 * u.Kelvin
    data = {
        "density": (np.ones(dims), "g/cm**3"),
        "temperature": (T_0.v * np.ones(dims), "K"),
        "velocity_x": (np.zeros(dims), "cm/s"),
        "velocity_y": (np.zeros(dims), "cm/s"),
        "velocity_z": (
            np.random.normal(loc=v_shift.v, scale=sigma_v.v, size=dims),
            "cm/s",
        ),
    }

    ds = load_uniform_grid(data, dims)

    cube = PPVCube(
        ds,
        "z",
        ("stream", "density"),
        (-300.0, 300.0, 1024, "km/s"),
        dims=8,
        thermal_broad=True,
    )

    dv = cube.dv
    v_th = np.sqrt(2.0 * kboltz * T_0 / (56.0 * mh) + 2.0 * sigma_v ** 2).in_units(
        "km/s"
    )
    a = cube.data.mean(axis=(0, 1)).v
    b = dv * np.exp(-(((cube.vmid + v_shift) / v_th) ** 2)) / (np.sqrt(np.pi) * v_th)

    assert_allclose_units(a, b, 1.0e-2)

    E_0 = 6.8 * u.keV

    cube.transform_spectral_axis(E_0.v, str(E_0.units))

    dE = -cube.dv
    delta_E = E_0 * v_th.in_cgs() / clight
    E_shift = E_0 * (1.0 + v_shift / clight)

    c = (
        dE
        * np.exp(-(((cube.vmid - E_shift) / delta_E) ** 2))
        / (np.sqrt(np.pi) * delta_E)
    )

    assert_allclose_units(a, c, 1.0e-2)


def test_ppv_nothermalbroad():

    np.random.seed(seed=0x4D3D3D3)

    dims = (16, 16, 128)
    v_shift = 1.0e6 * u.cm / u.s
    sigma_v = 2.0e6 * u.cm / u.s
    data = {
        "density": (np.ones(dims), "g/cm**3"),
        "velocity_x": (np.zeros(dims), "cm/s"),
        "velocity_y": (np.zeros(dims), "cm/s"),
        "velocity_z": (
            np.random.normal(loc=v_shift.v, scale=sigma_v.v, size=dims),
            "cm/s",
        ),
    }

    ds = load_uniform_grid(data, dims)

    cube = PPVCube(
        ds,
        "z",
        ("stream", "density"),
        (-100.0, 100.0, 128, "km/s"),
        dims=16,
        thermal_broad=False,
    )

    dv = cube.dv
    v_noth = np.sqrt(2) * (sigma_v).in_units("km/s")
    a = cube.data.mean(axis=(0, 1)).v
    b = (
        dv
        * np.exp(-(((cube.vmid + v_shift) / v_noth) ** 2))
        / (np.sqrt(np.pi) * v_noth)
    )

    assert_allclose_units(a, b, atol=5.0e-3)



"""
Unit test for the RADMC3D Exporter analysis module
"""

# -----------------------------------------------------------------------------
# Copyright (c) 2014, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

import os
import shutil
import tempfile

import numpy as np

import yt
from yt.testing import assert_allclose
from yt.utilities.answer_testing.framework import AnswerTestingTest, requires_ds
from yt_astro_analysis.radmc3d_export.api import RadMC3DWriter


class RadMC3DValuesTest(AnswerTestingTest):
    """

    This test writes out a "dust_density.inp" file,
    reads it back in, and checks the sum of the
    values for degradation.

    """

    _type_name = "RadMC3DValuesTest"
    _attrs = ("field",)

    def __init__(self, ds_fn, field, decimals=10):
        super().__init__(ds_fn)
        self.field = field
        self.decimals = decimals

    def run(self):

        # Set up in a temp dir
        tmpdir = tempfile.mkdtemp()
        curdir = os.getcwd()
        os.chdir(tmpdir)

        # try to write the output files
        writer = RadMC3DWriter(self.ds)
        writer.write_amr_grid()
        writer.write_dust_file(self.field, "dust_density.inp")

        # compute the sum of the values in the resulting file
        total = 0.0
        with open("dust_density.inp") as f:
            for i, line in enumerate(f):

                # skip header
                if i < 3:
                    continue

                line = line.rstrip()
                total += np.float64(line)

        # clean up
        os.chdir(curdir)
        shutil.rmtree(tmpdir)

        return total

    def compare(self, new_result, old_result):
        err_msg = f"Total value for {self.field} not equal."
        assert_allclose(
            new_result,
            old_result,
            10.0 ** (-self.decimals),
            err_msg=err_msg,
            verbose=True,
        )


etiny = "enzo_tiny_cosmology/DD0046/DD0046"


@requires_ds(etiny)
def test_radmc3d_exporter_continuum():
    """
    This test is simply following the description in the docs for how to
    generate the necessary output files to run a continuum emission map from
    dust for one of our sample datasets.
    """

    ds = yt.load(etiny)

    # Make up a dust density field where dust density is 1% of gas density
    dust_to_gas = 0.01

    def _DustDensity(field, data):
        return dust_to_gas * data["density"]

    ds.add_field(
        ("gas", "dust_density"),
        function=_DustDensity,
        sampling_type="cell",
        units="g/cm**3",
    )

    yield RadMC3DValuesTest(ds, ("gas", "dust_density"))



import os
import sys

from yt.frontends.halo_catalog.data_structures import YTHaloCatalogDataset
from yt.frontends.rockstar.data_structures import RockstarDataset
from yt.loaders import load
from yt_astro_analysis.utilities.testing import TempDirTest

_fields = (
    ("halos", "particle_position_x"),
    ("halos", "particle_position_y"),
    ("halos", "particle_position_z"),
    ("halos", "particle_mass"),
)

methods = {"fof": 2, "hop": 2, "rockstar": 3}
decimals = {"fof": 10, "hop": 10, "rockstar": 1}

etiny = "enzo_tiny_cosmology/32Mpc_32.enzo"


class HaloFinderTimeSeriesTest(TempDirTest):
    def test_halo_finders(self):
        from mpi4py import MPI

        filename = os.path.join(os.path.dirname(__file__), "run_halo_finder_ts.py")
        for method in methods:
            comm = MPI.COMM_SELF.Spawn(
                sys.executable,
                args=[filename, method, self.tmpdir],
                maxprocs=methods[method],
            )
            comm.Disconnect()

            if method == "rockstar":
                hcfns = [f"halos_{i}.0.bin" for i in range(2)]
            else:
                hcfns = [
                    os.path.join(f"DD{i:04d}", f"DD{i:04d}.0.h5") for i in [20, 46]
                ]

            for hcfn in hcfns:
                fn = os.path.join(self.tmpdir, "halo_catalogs", method, hcfn)

                ds = load(fn)
                if method == "rockstar":
                    ds.parameters["format_revision"] = 2
                    ds_type = RockstarDataset
                else:
                    ds_type = YTHaloCatalogDataset
                assert isinstance(ds, ds_type)



"""
testing utilities



"""

# -----------------------------------------------------------------------------
# Copyright (c) 2017, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

import os
import shutil
import tempfile
from unittest import TestCase

from yt.config import ytcfg
from yt.data_objects.time_series import SimulationTimeSeries
from yt.loaders import load_simulation
from yt.utilities.answer_testing.framework import AnswerTestingTest


class TempDirTest(TestCase):
    """
    A test class that runs in a temporary directory and
    removes it afterward.
    """

    def setUp(self):
        self.curdir = os.getcwd()
        self.tmpdir = tempfile.mkdtemp()
        os.chdir(self.tmpdir)

    def tearDown(self):
        os.chdir(self.curdir)
        shutil.rmtree(self.tmpdir)


def requires_sim(sim_fn, sim_type, file_check=False):
    from functools import wraps

    from nose import SkipTest

    def ffalse(func):
        @wraps(func)
        def fskip(*args, **kwargs):
            raise SkipTest

        return fskip

    def ftrue(func):
        return func

    if not can_run_sim(sim_fn, sim_type, file_check):
        return ffalse
    else:
        return ftrue


def can_run_sim(sim_fn, sim_type, file_check=False):
    result_storage = AnswerTestingTest.result_storage
    if isinstance(sim_fn, SimulationTimeSeries):
        return result_storage is not None
    path = ytcfg.get("yt", "test_data_dir")
    if not os.path.isdir(path):
        return False
    if file_check:
        return os.path.isfile(os.path.join(path, sim_fn)) and result_storage is not None
    try:
        load_simulation(sim_fn, sim_type)
    except FileNotFoundError:
        if ytcfg.get("yt", "internals", "strict_requires"):
            if result_storage is not None:
                result_storage["tainted"] = True
            raise
        return False
    return result_storage is not None



"""
HaloCatalog recipes



"""

from yt_astro_analysis.halo_analysis.halo_catalog.analysis_operators import add_recipe


def calculate_virial_quantities(
    pipeline,
    fields,
    weight_field=None,
    accumulation=True,
    radius_field="virial_radius",
    factor=2.0,
    overdensity_field=("gas", "overdensity"),
    critical_overdensity=200,
):
    r"""
    Calculate virial quantities with the following procedure:
    1. Create a sphere data container.
    2. Create 1D radial profiles of overdensity and any requested fields.
    3. Call virial_quantities callback to interpolate profiles for value of critical overdensity.
    4. Delete profile and sphere objects from halo.

    Parameters
    ----------
    halo : Halo object
        The Halo object to be provided by the HaloCatalog.
    fields: string or list of strings
        The fields for which virial values are to be calculated.
    weight_field : string
        Weight field for profiling.
        Default : "cell_mass"
    accumulation : bool or list of bools
        If True, the profile values for a bin n are the cumulative sum of
        all the values from bin 0 to n.  If -True, the sum is reversed so
        that the value for bin n is the cumulative sum from bin N (total bins)
        to n.  If the profile is 2D or 3D, a list of values can be given to
        control the summation in each dimension independently.
        Default: False.
    radius_field : string
        Field to be retrieved from the quantities dictionary as
        the basis of the halo radius.
        Default: "virial_radius".
    factor : float
        Factor to be multiplied by the base radius for defining
        the radius of the sphere.
        Default: 2.0.
    overdensity_field : string or tuple of strings
        The field used as the overdensity from which interpolation is done to
        calculate virial quantities.
        Default: ("gas", "overdensity")
    critical_overdensity : float
        The value of the overdensity at which to evaluate the virial quantities.
        Overdensity is with respect to the critical density.
        Default: 200

    """

    storage = "virial_quantities_profiles"
    pfields = [field for field in fields if field != "radius"]

    pipeline.add_callback("sphere", factor=factor)
    if pfields:
        pipeline.add_callback(
            "profile",
            ["radius"],
            pfields,
            weight_field=weight_field,
            accumulation=accumulation,
            storage=storage,
        )
    pipeline.add_callback(
        "profile",
        ["radius"],
        [overdensity_field],
        weight_field="cell_volume",
        accumulation=True,
        storage=storage,
    )
    pipeline.add_callback(
        "virial_quantities",
        fields,
        overdensity_field=overdensity_field,
        critical_overdensity=critical_overdensity,
        profile_storage=storage,
    )
    pipeline.add_callback("delete_attribute", storage)
    pipeline.add_callback("delete_attribute", "data_object")


add_recipe("calculate_virial_quantities", calculate_virial_quantities)



import yt_astro_analysis.halo_analysis.halo_catalog.plot_modifications
from yt_astro_analysis.halo_analysis.halo_catalog.analysis_operators import add_quantity
from yt_astro_analysis.halo_analysis.halo_catalog.halo_callbacks import add_callback
from yt_astro_analysis.halo_analysis.halo_catalog.halo_catalog import HaloCatalog
from yt_astro_analysis.halo_analysis.halo_catalog.halo_filters import add_filter
from yt_astro_analysis.halo_analysis.halo_catalog.halo_finding_methods import (
    add_finding_method,
)
from yt_astro_analysis.halo_analysis.halo_catalog.halo_recipes import add_recipe



"""
halo finding



"""

# -----------------------------------------------------------------------------
# Copyright (c) yt Development Team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

import gc

import numpy as np

from yt.config import ytcfg
from yt.funcs import mylog
from yt.utilities.math_utils import get_rotation_matrix, periodic_dist
from yt.utilities.parallel_tools.parallel_analysis_interface import (
    ParallelAnalysisInterface,
)
from yt.utilities.physical_constants import mass_sun_cgs
from yt.utilities.physical_ratios import TINY, rho_crit_g_cm3_h2
from yt_astro_analysis.halo_analysis.halo_finding.fof.EnzoFOF import RunFOF
from yt_astro_analysis.halo_analysis.halo_finding.hop.EnzoHop import RunHOP


class Halo:
    """
    A data source that returns particle information about the members of a
    HOP-identified halo.
    """

    _distributed = False
    _processing = False
    _owner = 0
    indices = None
    extra_wrap = ["__getitem__"]

    def __init__(
        self,
        halo_list,
        id,
        indices=None,
        size=None,
        CoM=None,
        max_dens_point=None,
        group_total_mass=None,
        max_radius=None,
        bulk_vel=None,
        tasks=None,
        rms_vel=None,
        supp=None,
        ptype="all",
    ):
        self.ptype = ptype
        self.halo_list = halo_list
        self._max_dens = halo_list._max_dens
        self.id = id
        self.data = halo_list._data_source
        self.ds = self.data.ds
        self.gridsize = self.ds.domain_right_edge - self.ds.domain_left_edge
        if indices is not None:
            self.indices = halo_list._base_indices[indices]
        else:
            self.indices = None
        # We assume that if indices = None, the instantiator has OTHER plans
        # for us -- i.e., setting it somehow else
        self.size = size
        self.CoM = CoM
        self.max_dens_point = max_dens_point
        self.group_total_mass = group_total_mass
        self.max_radius = max_radius
        self.bulk_vel = bulk_vel
        self.tasks = tasks
        self.rms_vel = rms_vel
        self.bin_count = None
        self.overdensity = None
        # A supplementary data dict.
        if supp is None:
            self.supp = {}
        else:
            self.supp = supp
        self._saved_fields = {}
        self._ds_sort = None
        self._particle_mask = None

    @property
    def particle_mask(self):
        # Dynamically create the masking array for particles, and get
        # the data using standard yt methods.
        if self._particle_mask is not None:
            return self._particle_mask
        # This is from disk.
        pid = self.__getitem__("particle_index")
        # This is from the sphere.
        if self._name == "RockstarHalo":
            ds = self.ds.sphere(self.CoM, self._radjust * self.max_radius)
        elif self._name == "LoadedHalo":
            ds = self.ds.sphere(
                self.CoM,
                np.maximum(
                    self._radjust * self.ds.quan(self.max_radius, "code_length"),
                    self.ds.index.get_smallest_dx(),
                ),
            )
        sp_pid = ds["particle_index"]
        self._ds_sort = sp_pid.argsort()
        sp_pid = sp_pid[self._ds_sort]
        # This matches them up.
        self._particle_mask = np.in1d(sp_pid, pid)
        return self._particle_mask

    def center_of_mass(self):
        r"""Calculate and return the center of mass.

        The center of mass of the halo is directly calculated and returned.

        Examples
        --------
        >>> com = halos[0].center_of_mass()
        """
        if self.CoM is not None:
            return self.CoM
        pm = self["particle_mass"].in_units("Msun")
        c = {}
        # We shift into a box where the origin is the left edge
        c[0] = self["particle_position_x"] - self.ds.domain_left_edge[0]
        c[1] = self["particle_position_y"] - self.ds.domain_left_edge[1]
        c[2] = self["particle_position_z"] - self.ds.domain_left_edge[2]
        com = []
        for i in range(3):
            # A halo is likely periodic around a boundary if the distance
            # between the max and min particle
            # positions are larger than half the box.
            # So skip the rest if the converse is true.
            # Note we might make a change here when periodicity-handling is
            # fully implemented.
            if (c[i].max() - c[i].min()) < (self.ds.domain_width[i] / 2.0):
                com.append(c[i])
                continue
            # Now we want to flip around only those close to the left boundary.
            sel = c[i] <= (self.ds.domain_width[i] / 2)
            c[i][sel] += self.ds.domain_width[i]
            com.append(c[i])

        c = (com * pm).sum(axis=1) / pm.sum()
        c = self.ds.arr(c, "code_length")

        return c % self.ds.domain_width + self.ds.domain_left_edge

    def maximum_density(self):
        r"""Return the HOP-identified maximum density. Not applicable to
        FOF halos.

        Return the HOP-identified maximum density. Not applicable to FOF halos.

        Examples
        --------
        >>> max_dens = halos[0].maximum_density()
        """
        if self.max_dens_point is not None:
            return self.max_dens_point[0]
        return self._max_dens[self.id][0]

    def maximum_density_location(self):
        r"""Return the location HOP identified as maximally dense. Not
        applicable to FOF halos.

        Return the location HOP identified as maximally dense.

        Examples
        --------
        >>> max_dens_loc = halos[0].maximum_density_location()
        """
        if self.max_dens_point is not None:
            return self.max_dens_point[1:]
        return np.array(
            [
                self._max_dens[self.id][1],
                self._max_dens[self.id][2],
                self._max_dens[self.id][3],
            ]
        )

    def total_mass(self):
        r"""Returns the total mass in solar masses of the halo.

        Returns the total mass in solar masses of just the particles in the
        halo.

        Examples
        --------
        >>> halos[0].total_mass()
        """
        if self.group_total_mass is not None:
            return self.group_total_mass
        return self["particle_mass"].in_units("Msun").sum()

    def bulk_velocity(self):
        r"""Returns the mass-weighted average velocity in cm/s.

        This calculates and returns the mass-weighted average velocity of just
        the particles in the halo in cm/s.

        Examples
        --------
        >>> bv = halos[0].bulk_velocity()
        """
        if self.bulk_vel is not None:
            return self.bulk_vel
        pm = self["particle_mass"].in_units("Msun")
        vx = (self["particle_velocity_x"] * pm).sum()
        vy = (self["particle_velocity_y"] * pm).sum()
        vz = (self["particle_velocity_z"] * pm).sum()
        return self.ds.arr([vx, vy, vz], vx.units) / pm.sum()

    def rms_velocity(self):
        r"""Returns the mass-weighted RMS velocity for the halo
        particles in cgs units.

        Calculate and return the mass-weighted RMS velocity for just the
        particles in the halo.  The bulk velocity of the halo is subtracted
        before computation.

        Examples
        --------
        >>> rms_vel = halos[0].rms_velocity()
        """
        if self.rms_vel is not None:
            return self.rms_vel
        bv = self.bulk_velocity()
        pm = self["particle_mass"].in_units("Msun")
        sm = pm.sum()
        vx = (self["particle_velocity_x"] - bv[0]) * pm / sm
        vy = (self["particle_velocity_y"] - bv[1]) * pm / sm
        vz = (self["particle_velocity_z"] - bv[2]) * pm / sm
        s = vx ** 2.0 + vy ** 2.0 + vz ** 2.0
        ms = np.mean(s)
        return np.sqrt(ms) * pm.size

    def maximum_radius(self, center_of_mass=True):
        r"""Returns the maximum radius in the halo for all particles,
        either from the point of maximum density or from the
        center of mass.

        The maximum radius from the most dense point is calculated.  This
        accounts for periodicity.

        Parameters
        ----------
        center_of_mass : bool
            True chooses the center of mass when
            calculating the maximum radius.
            False chooses from the maximum density location for HOP halos
            (it has no effect for FOF halos).
            Default = True.

        Examples
        --------
        >>> radius = halos[0].maximum_radius()
        """
        if self.max_radius is not None:
            return self.max_radius
        if center_of_mass:
            center = self.center_of_mass()
        else:
            center = self.maximum_density_location()
        rx = np.abs(self["particle_position_x"] - center[0])
        ry = np.abs(self["particle_position_y"] - center[1])
        rz = np.abs(self["particle_position_z"] - center[2])
        DW = self.data.ds.domain_right_edge - self.data.ds.domain_left_edge
        r = np.sqrt(
            np.minimum(rx, DW[0] - rx) ** 2.0
            + np.minimum(ry, DW[1] - ry) ** 2.0
            + np.minimum(rz, DW[2] - rz) ** 2.0
        )
        return r.max()

    def __getitem__(self, key):
        return self.data[(self.ptype, key)][self.indices]

    def get_sphere(self, center_of_mass=True):
        r"""Returns a sphere source.

        This will generate a new, empty sphere source centered on this halo,
        with the maximum radius of the halo. This can be used like any other
        data container in yt.

        Parameters
        ----------
        center_of_mass : bool, optional
            True chooses the center of mass when
            calculating the maximum radius.
            False chooses from the maximum density location for HOP halos
            (it has no effect for FOF halos).
            Default = True.

        Returns
        -------
        sphere : `yt.data_objects.api.YTSphere`
            The empty data source.

        Examples
        --------
        >>> sp = halos[0].get_sphere()
        """
        if center_of_mass:
            center = self.center_of_mass()
        else:
            center = self.maximum_density_location()
        radius = self.maximum_radius()
        # A bit of a long-reach here...
        sphere = self.data.ds.sphere(center, radius=radius)
        return sphere

    def get_size(self):
        if self.size is not None:
            return self.size
        return self.indices.size

    def virial_mass(self, virial_overdensity=200.0, bins=300):
        r"""Return the virial mass of the halo in Msun,
        using only the particles
        in the halo (no baryonic information used).

        The virial mass is calculated, using the built in `Halo.virial_info`
        functionality.  The mass is then returned.

        Parameters
        ----------
        virial_overdensity : float
            The overdensity threshold compared to the universal average when
            calculating the virial mass. Default = 200.
        bins : int
            The number of spherical bins used to calculate overdensities.
            Default = 300.

        Returns
        -------
        mass : float
            The virial mass in solar masses of the particles in the halo.  -1
            if not virialized.

        Examples
        --------
        >>> vm = halos[0].virial_mass()
        """
        self.virial_info(bins=bins)
        vir_bin = self.virial_bin(virial_overdensity=virial_overdensity, bins=bins)
        if vir_bin != -1:
            return self.mass_bins[vir_bin]
        else:
            return -1

    def virial_radius(self, virial_overdensity=200.0, bins=300):
        r"""Return the virial radius of the halo in code units.

        The virial radius of the halo is calculated, using only the particles
        in the halo (no baryonic information used). Returns -1 if the halo is
        not virialized.

        Parameters
        ----------
        virial_overdensity : float
            The overdensity threshold compared to the universal average when
            calculating the virial radius. Default = 200.
        bins : integer
            The number of spherical bins used to calculate overdensities.
            Default = 300.

        Returns
        -------
        radius : float
            The virial radius in code units of the particles in the halo.  -1
            if not virialized.

        Examples
        --------
        >>> vr = halos[0].virial_radius()
        """
        self.virial_info(bins=bins)
        vir_bin = self.virial_bin(virial_overdensity=virial_overdensity, bins=bins)
        if vir_bin != -1:
            return self.radial_bins[vir_bin]
        else:
            return -1

    def virial_bin(self, virial_overdensity=200.0, bins=300):
        r"""Returns the bin index of the virial radius of the halo. Generally,
        it is better to call virial_radius instead, which calls this function
        automatically.
        """
        self.virial_info(bins=bins)
        over = self.overdensity > virial_overdensity
        if over.any():
            vir_bin = max(np.arange(bins + 1)[over])
            return vir_bin
        else:
            return -1

    def virial_info(self, bins=300):
        r"""Calculates the virial information for the halo. Generally, it is
        better to call virial_radius or virial_mass instead, which calls this
        function automatically.
        """
        # Skip if we've already calculated for this number of bins.
        if self.bin_count == bins and self.overdensity is not None:
            return None
        self.bin_count = bins
        # Cosmology
        h = self.ds.hubble_constant
        Om_matter = self.ds.omega_matter
        z = self.ds.current_redshift
        period = self.ds.domain_right_edge - self.ds.domain_left_edge
        thissize = self.get_size()
        rho_crit = rho_crit_g_cm3_h2 * h ** 2.0 * Om_matter  # g cm^-3
        Msun2g = mass_sun_cgs
        rho_crit = rho_crit * ((1.0 + z) ** 3.0)
        # Get some pertinent information about the halo.
        self.mass_bins = self.ds.arr(
            np.zeros(self.bin_count + 1, dtype="float64"), "Msun"
        )
        dist = np.empty(thissize, dtype="float64")
        cen = self.center_of_mass()
        mark = 0
        # Find the distances to the particles. I don't like this much, but I
        # can't see a way to eliminate a loop like this, either here or in yt.
        for pos in zip(
            self["particle_position_x"],
            self["particle_position_y"],
            self["particle_position_z"],
        ):
            dist[mark] = periodic_dist(cen, pos, period)
            mark += 1
        # Set up the radial bins.
        # Multiply min and max to prevent issues with digitize below.
        self.radial_bins = np.logspace(
            np.log10(min(dist) * 0.99 + TINY),
            np.log10(max(dist) * 1.01 + 2 * TINY),
            num=self.bin_count + 1,
        )
        self.radial_bins = self.ds.arr(self.radial_bins, "code_length")
        # Find out which bin each particle goes into, and add the particle
        # mass to that bin.
        inds = np.digitize(dist, self.radial_bins) - 1
        if self["particle_position_x"].size > 1:
            for index in np.unique(inds):
                self.mass_bins[index] += np.sum(
                    self["particle_mass"][inds == index]
                ).in_units("Msun")
        # Now forward sum the masses in the bins.
        for i in range(self.bin_count):
            self.mass_bins[i + 1] += self.mass_bins[i]
        # Calculate the over densities in the bins.
        self.overdensity = (
            self.mass_bins
            * Msun2g
            / (4.0 / 3.0 * np.pi * rho_crit * (self.radial_bins) ** 3.0)
        )

    def _get_ellipsoid_parameters_basic(self):
        np.seterr(all="ignore")
        # check if there are 4 particles to form an ellipsoid
        # neglecting to check if the 4 particles in the same plane,
        # that is almost certainly never to occur,
        # will deal with it later if it ever comes up
        if np.size(self["particle_position_x"]) < 4:
            mylog.warning("Too few particles for ellipsoid parameters.")
            return (0, 0, 0, 0, 0, 0, 0)
        # Calculate the parameters that describe the ellipsoid of
        # the particles that constitute the halo. This function returns
        # all the parameters except for the center of mass.
        com = self.center_of_mass()
        position = [
            self["particle_position_x"],
            self["particle_position_y"],
            self["particle_position_z"],
        ]
        # Locate the furthest particle from com, its vector length and index
        DW = np.array([self.gridsize[0], self.gridsize[1], self.gridsize[2]])
        position = [position[0] - com[0], position[1] - com[1], position[2] - com[2]]
        # different cases of particles being on other side of boundary
        for axis in range(np.size(DW)):
            cases = np.array(
                [position[axis], position[axis] + DW[axis], position[axis] - DW[axis]]
            )
            # pick out the smallest absolute distance from com
            position[axis] = np.choose(np.abs(cases).argmin(axis=0), cases)
        # find the furthest particle's index
        r = np.sqrt(position[0] ** 2 + position[1] ** 2 + position[2] ** 2)
        A_index = r.argmax()
        mag_A = r.max()
        # designate the A vector
        A_vector = (position[0][A_index], position[1][A_index], position[2][A_index])
        # designate the e0 unit vector
        e0_vector = A_vector / mag_A
        # locate the tB particle position by finding the max B
        e0_vector_copy = np.empty((np.size(position[0]), 3), dtype="float64")
        for i in range(3):
            e0_vector_copy[:, i] = e0_vector[i]
        rr = np.array(
            [position[0], position[1], position[2]]
        ).T  # Similar to tB_vector in old code.
        tC_vector = np.cross(e0_vector_copy, rr)
        te2 = tC_vector.copy()
        for dim in range(3):
            te2[:, dim] *= np.sum(tC_vector ** 2.0, axis=1) ** (-0.5)
        te1 = np.cross(te2, e0_vector_copy)
        length = np.abs(
            -np.sum(rr * te1, axis=1)
            * (1.0 - np.sum(rr * e0_vector_copy, axis=1) ** 2.0 * mag_A ** -2.0)
            ** (-0.5)
        )
        # This problem apparently happens sometimes, that the NaNs are turned
        # into infs, which messes up the nanargmax below.
        length[length == np.inf] = 0.0
        tB_index = np.nanargmax(length)  # ignores NaNs created above.
        mag_B = length[tB_index]
        e1_vector = te1[tB_index]
        e2_vector = te2[tB_index]
        temp_e0 = rr.copy()
        temp_e1 = rr.copy()
        temp_e2 = rr.copy()
        for dim in range(3):
            temp_e0[:, dim] = e0_vector[dim]
            temp_e1[:, dim] = e1_vector[dim]
            temp_e2[:, dim] = e2_vector[dim]
        length = np.abs(
            np.sum(rr * temp_e2, axis=1)
            * (
                1
                - np.sum(rr * temp_e0, axis=1) ** 2.0 * mag_A ** -2.0
                - np.sum(rr * temp_e1, axis=1) ** 2.0 * mag_B ** -2.0
            )
            ** (-0.5)
        )
        length[length == np.inf] = 0.0
        tC_index = np.nanargmax(length)
        mag_C = length[tC_index]
        # tilt is calculated from the rotation about x axis
        # needed to align e1 vector with the y axis
        # after e0 is aligned with x axis
        # find the t1 angle needed to rotate about z axis to align e0 onto x-z plane
        t1 = np.arctan(-e0_vector[1] / e0_vector[0])
        RZ = get_rotation_matrix(t1, (0, 0, 1))
        r1 = np.dot(RZ, e0_vector)
        # find the t2 angle needed to rotate about y axis to align e0 to x
        t2 = np.arctan(r1[2] / r1[0])
        RY = get_rotation_matrix(t2, (0, 1, 0))
        r2 = np.dot(RY, np.dot(RZ, e1_vector))
        # find the tilt angle needed to rotate about x axis to align e1 to y and e2 to z
        tilt = np.arctan(-r2[2] / r2[1])
        return (mag_A, mag_B, mag_C, e0_vector[0], e0_vector[1], e0_vector[2], tilt)


class HOPHalo(Halo):
    _name = "HOPHalo"
    pass


class FOFHalo(Halo):
    def maximum_density(self):
        r"""Not implemented."""
        return -1

    def maximum_density_location(self):
        r"""Not implemented."""
        return self.center_of_mass()


class HaloList:

    _fields = ["particle_position_%s" % ax for ax in "xyz"]

    def __init__(self, data_source, redshift=-1, ptype="all"):
        """
        Run hop on *data_source* with a given density *threshold*.
        Returns an iterable collection of
        *HopGroup* items.
        """
        self._data_source = data_source
        self.ptype = ptype
        self._groups = []
        self._max_dens = {}
        self.__obtain_particles()
        self._run_finder()
        mylog.info("Parsing outputs")
        self._parse_output()
        mylog.debug("Finished. (%s)", len(self))
        self.redshift = redshift

    def __obtain_particles(self):
        ii = slice(None)
        self.particle_fields = {}
        for field in self._fields:
            tot_part = self._data_source[(self.ptype, field)].size
            if field == "particle_index":
                self.particle_fields[field] = self._data_source[(self.ptype, field)][
                    ii
                ].astype("int64")
            else:
                self.particle_fields[field] = self._data_source[(self.ptype, field)][
                    ii
                ].astype("float64")
            del self._data_source[(self.ptype, field)]
        self._base_indices = np.arange(tot_part)[ii]
        gc.collect()

    def _parse_output(self):
        unique_ids = np.unique(self.tags)
        counts = np.bincount(self.tags + 1)
        sort_indices = np.argsort(self.tags)
        grab_indices = np.indices(self.tags.shape).ravel()[sort_indices]
        dens = self.densities[sort_indices]
        cp = 0
        for i in unique_ids:
            cp_c = cp + counts[i + 1]
            if i == -1:
                cp += counts[i + 1]
                continue
            group_indices = grab_indices[cp:cp_c]
            self._groups.append(
                self._halo_class(self, i, group_indices, ptype=self.ptype)
            )
            md_i = np.argmax(dens[cp:cp_c])
            px, py, pz = (
                self.particle_fields["particle_position_%s" % ax][group_indices]
                for ax in "xyz"
            )
            self._max_dens[i] = (dens[cp:cp_c][md_i], px[md_i], py[md_i], pz[md_i])
            cp += counts[i + 1]

    def __len__(self):
        return len(self._groups)

    def __iter__(self):
        yield from self._groups

    def __getitem__(self, key):
        return self._groups[key]


class HOPHaloList(HaloList):
    """
    Run hop on *data_source* with a given density *threshold*.
    Returns an iterable collection of *HopGroup* items.
    """

    _name = "HOP"
    _halo_class = HOPHalo
    _fields = ["particle_position_%s" % ax for ax in "xyz"] + ["particle_mass"]

    def __init__(self, data_source, threshold=160.0, ptype="all"):
        self.threshold = threshold
        mylog.info("Initializing HOP")
        HaloList.__init__(self, data_source, ptype=ptype)

    def _run_finder(self):
        self.densities, self.tags = RunHOP(
            self.particle_fields["particle_position_x"] / self.period[0],
            self.particle_fields["particle_position_y"] / self.period[1],
            self.particle_fields["particle_position_z"] / self.period[2],
            self.particle_fields["particle_mass"].in_units("Msun"),
            self.threshold,
        )
        self.particle_fields["densities"] = self.densities
        self.particle_fields["tags"] = self.tags


class FOFHaloList(HaloList):
    _name = "FOF"
    _halo_class = FOFHalo

    def __init__(self, data_source, link=0.2, redshift=-1, ptype="all"):
        self.link = link
        mylog.info("Initializing FOF")
        HaloList.__init__(self, data_source, redshift=redshift, ptype=ptype)

    def _run_finder(self):
        self.tags = RunFOF(
            self.particle_fields["particle_position_x"] / self.period[0],
            self.particle_fields["particle_position_y"] / self.period[1],
            self.particle_fields["particle_position_z"] / self.period[2],
            self.link,
        )
        self.densities = np.ones(self.tags.size, dtype="float64") * -1
        self.particle_fields["densities"] = self.densities
        self.particle_fields["tags"] = self.tags


class GenericHaloFinder(HaloList, ParallelAnalysisInterface):
    def __init__(self, ds, data_source, padding=0.0, ptype="all"):
        ParallelAnalysisInterface.__init__(self)
        self.ds = ds
        self.index = ds.index
        self.center = (
            np.array(data_source.right_edge) + np.array(data_source.left_edge)
        ) / 2.0
        self.ptype = ptype

    def _parse_halolist(self, threshold_adjustment):
        groups = []
        max_dens = {}
        hi = 0
        LE, RE = self.bounds
        for halo in self._groups:
            this_max_dens = halo.maximum_density_location()
            # if the most dense particle is in the box, keep it
            if np.all((this_max_dens >= LE) & (this_max_dens <= RE)):
                # Now we add the halo information to OURSELVES, taken from the
                # self.hop_list
                # We need to mock up the HOPHaloList thingie, so we need to
                #     set self._max_dens
                max_dens_temp = list(self._max_dens[halo.id])[0] / threshold_adjustment
                max_dens[hi] = [max_dens_temp] + list(self._max_dens[halo.id])[1:4]
                groups.append(self._halo_class(self, hi, ptype=self.ptype))
                groups[-1].indices = halo.indices
                self.comm.claim_object(groups[-1])
                hi += 1
        del self._groups, self._max_dens  # explicit >> implicit
        self._groups = groups
        self._max_dens = max_dens

    def _join_halolists(self):
        groups = {self.comm.rank: len(self)}
        groups = self.comm.par_combine_object(groups, datatype="dict", op="join")

        ngroups = np.array([groups[rank] for rank in sorted(groups)])
        offsets = ngroups.cumsum() - ngroups
        my_offset = offsets[self.comm.rank]
        for halo in self:
            halo.id += my_offset

    def _reposition_particles(self, bounds):
        # This only does periodicity.  We do NOT want to deal with anything
        # else.  The only reason we even do periodicity is the
        LE, RE = bounds
        dw = self.ds.domain_right_edge - self.ds.domain_left_edge
        for i, ax in enumerate("xyz"):
            arr = self._data_source[self.ptype, "particle_position_%s" % ax]
            arr[arr < LE[i] - self.padding] += dw[i]
            arr[arr > RE[i] + self.padding] -= dw[i]


class HOPHaloFinder(GenericHaloFinder, HOPHaloList):
    r"""HOP halo finder.

    Halos are built by:
    1. Calculating a density for each particle based on a smoothing kernel.
    2. Recursively linking particles to other particles from lower density
    particles to higher.
    3. Geometrically proximate chains are identified and
    4. merged into final halos following merging rules.

    Lower thresholds generally produce more halos, and the largest halos
    become larger. Also, halos become more filamentary and over-connected.

    Eisenstein and Hut. "HOP: A New Group-Finding Algorithm for N-Body
    Simulations." ApJ (1998) vol. 498 pp. 137-142

    Parameters
    ----------
    ds : `Dataset`
        The dataset on which halo finding will be conducted.
    subvolume : `yt.data_objects.data_containers.YTSelectionContainer`, optional
        A region over which HOP will be run, which can be used to run HOP
        on a subvolume of the full volume. Default = None, which defaults
        to the full volume automatically.
    threshold : float
        The density threshold used when building halos. Default = 160.0.
    ptype : string
        The particle type to be used for halo finding.
        Default: 'all'.
    padding : float
        When run in parallel, the finder needs to surround each subvolume
        with duplicated particles for halo finidng to work. This number
        must be no smaller than the radius of the largest halo in the box
        in code units. Default = 0.02.
    total_mass : float
        If HOP is run on the same dataset mulitple times, the total mass
        of particles in Msun units in the full volume can be supplied here
        to save time.
        This must correspond to the particles being operated on, meaning
        if stars are included in the halo finding, they must be included
        in this mass as well, and visa-versa.
        If halo finding on a subvolume, this still corresponds with the
        mass in the entire volume.
        Default = None, which means the total mass is automatically
        calculated.
    save_particles : bool
        If True, output member particles for each halo.
        Default: True.

    Examples
    --------
    >>> import yt
    >>> from yt.extensions.astro_analysis.halo_analysis import HaloCatalog
    >>> data_ds = yt.load('Enzo_64/RD0006/RedshiftOutput0006')
    >>> hc = HaloCatalog(data_ds=data_ds, finder_method='hop',
    ...                  finder_kwargs={"threshold": 160})
    >>> hc.create()
    """

    def __init__(
        self,
        ds,
        subvolume=None,
        threshold=160,
        dm_only=False,
        ptype="all",
        padding=0.02,
        total_mass=None,
        save_particles=True,
    ):
        if subvolume is not None:
            ds_LE = np.array(subvolume.left_edge)
            ds_RE = np.array(subvolume.right_edge)
        self.period = ds.domain_right_edge - ds.domain_left_edge
        self.save_particles = save_particles
        self._data_source = ds.all_data()
        GenericHaloFinder.__init__(self, ds, self._data_source, padding, ptype=ptype)
        # do it once with no padding so the total_mass is correct
        # (no duplicated particles), and on the entire volume, even if only
        # a small part is actually going to be used.
        self.padding = 0.0
        padded, LE, RE, self._data_source = self.partition_index_3d(
            ds=self._data_source, padding=self.padding
        )

        if dm_only:
            raise RuntimeError(
                "dm_only has been removed. "
                + "Use ptype to specify a particle type, instead."
            )

        self.ptype = ptype

        # For scaling the threshold, note that it's a passthrough
        if total_mass is None:
            total_mass = self.comm.mpi_allreduce(
                self._data_source.quantities.total_quantity(
                    (self.ptype, "particle_mass")
                ).in_units("Msun"),
                op="sum",
            )
        # MJT: Note that instead of this, if we are assuming that the particles
        # are all on different processors, we should instead construct an
        # object representing the entire domain and sum it "lazily" with
        # Derived Quantities.
        if subvolume is not None:
            self._data_source = ds.region([0.0] * 3, ds_LE, ds_RE)
        else:
            self._data_source = ds.all_data()
        self.padding = padding  # * ds["unitary"] # This should be clevererer
        padded, LE, RE, self._data_source = self.partition_index_3d(
            ds=self._data_source, padding=self.padding
        )
        self.bounds = (LE, RE)
        # sub_mass can be skipped if subvolume is not used and this is not
        # parallel.
        if (
            subvolume is None
            and ytcfg.get("yt", "internals", "topcomm_parallel_size") == 1
        ):
            sub_mass = total_mass
        else:
            sub_mass = self._data_source.quantities.total_quantity(
                (self.ptype, "particle_mass")
            ).in_units("Msun")
        HOPHaloList.__init__(
            self, self._data_source, threshold * total_mass / sub_mass, ptype=self.ptype
        )
        self._parse_halolist(total_mass / sub_mass)
        self._join_halolists()


class FOFHaloFinder(GenericHaloFinder, FOFHaloList):
    r"""Friends-of-friends halo finder.

    Halos are found by linking together all pairs of particles closer than
    some distance from each other. Particles may have multiple links,
    and halos are found by recursively linking together all such pairs.

    Larger linking lengths produce more halos, and the largest halos
    become larger. Also, halos become more filamentary and over-connected.

    Davis et al. "The evolution of large-scale structure in a universe
    dominated by cold dark matter." ApJ (1985) vol. 292 pp. 371-394

    Parameters
    ----------
    ds : `Dataset`
        The dataset on which halo finding will be conducted.
    subvolume : `yt.data_objects.data_containers.YTSelectionContainer`, optional
        A region over which HOP will be run, which can be used to run HOP
        on a subvolume of the full volume. Default = None, which defaults
        to the full volume automatically.
    link : float
        If positive, the interparticle distance (compared to the overall
        average) used to build the halos. If negative, this is taken to be
        the *actual* linking length, and no other calculations will be
        applied.  Default = 0.2.
    ptype : string
        The type of particle to be used for halo finding.
        Default: 'all'.
    padding : float
        When run in parallel, the finder needs to surround each subvolume
        with duplicated particles for halo finidng to work. This number
        must be no smaller than the radius of the largest halo in the box
        in code units. Default = 0.02.
    save_particles : bool
        If True, output member particles for each halo.
        Default: True.

    Examples
    --------
    >>> import yt
    >>> from yt.extensions.astro_analysis.halo_analysis import HaloCatalog
    >>> data_ds = yt.load('Enzo_64/RD0006/RedshiftOutput0006')
    >>> hc = HaloCatalog(data_ds=data_ds, finder_method='fof',
    ...                  finder_kwargs={"link": 0.2})
    >>> hc.create()
    """

    def __init__(
        self,
        ds,
        subvolume=None,
        link=0.2,
        dm_only=False,
        ptype="all",
        padding=0.02,
        save_particles=True,
    ):
        if subvolume is not None:
            ds_LE = np.array(subvolume.left_edge)
            ds_RE = np.array(subvolume.right_edge)
        self.period = ds.domain_right_edge - ds.domain_left_edge
        self.ds = ds
        self.index = ds.index
        self.redshift = ds.current_redshift
        self.save_particles = save_particles
        self._data_source = ds.all_data()
        GenericHaloFinder.__init__(self, ds, self._data_source, padding)
        self.padding = 0.0  # * ds["unitary"] # This should be clevererer
        # get the total number of particles across all procs, with no padding
        padded, LE, RE, self._data_source = self.partition_index_3d(
            ds=self._data_source, padding=self.padding
        )

        if dm_only:
            raise RuntimeError(
                "dm_only has been removed. "
                + "Use ptype to specify a particle type, instead."
            )

        self.ptype = ptype

        if link > 0.0:
            n_parts = self.comm.mpi_allreduce(
                self._data_source[self.ptype, "particle_ones"].size, op="sum"
            )
            # get the average spacing between particles
            # l = ds.domain_right_edge - ds.domain_left_edge
            # vol = l[0] * l[1] * l[2]
            # Because we are now allowing for datasets with non 1-periodicity,
            # but symmetric, vol is always 1.
            vol = 1.0
            avg_spacing = (float(vol) / n_parts) ** (1.0 / 3.0)
            linking_length = link * avg_spacing
        else:
            linking_length = np.abs(link)
        self.padding = padding
        if subvolume is not None:
            self._data_source = ds.region([0.0] * 3, ds_LE, ds_RE)
        else:
            self._data_source = ds.all_data()
        padded, LE, RE, self._data_source = self.partition_index_3d(
            ds=self._data_source, padding=self.padding
        )
        self.bounds = (LE, RE)
        # reflect particles around the periodic boundary
        # self._reposition_particles((LE, RE))
        # here is where the FOF halo finder is run
        mylog.info("Using a linking length of %0.3e", linking_length)
        FOFHaloList.__init__(
            self,
            self._data_source,
            linking_length,
            redshift=self.redshift,
            ptype=self.ptype,
        )
        self._parse_halolist(1.0)
        self._join_halolists()
